classdef BSolver < handle & matlab.mixin.CustomDisplay % & matlab.mixin.Copyable
    % Abstract class for estimating (fitting) base terms from a
    % :class:`Signal`. Derived classes, generally referred to as
    % *BSolvers*, must implement a:: 
    %
    %   b = BSolverSubclass.solve( self, signal )
    %
    % method that computes the normalized base terms from a :class:`Signal`
    % object. All BSolvers inherit class *BSolver*.
    %
    % Every *BSolver* object has a *parms* property of class
    % *Parameters* which contains at least a *nsamples* parameter
    % (*parms.p.nsamples.value*). 
    
    properties
        parms    % Parameters object with solver parameters
    end
          
    methods
        %------------------------------------------------------------------------------------------
        function init( self, varargin )
            % [delayed constructor] Initialize (or re-initialize) a
            % BSolver. This sets the solver parameter
            % self.parms.p.nsamples.
            %
            % The init method checks the validity of all parameters. The
            % validity of the parameter combination is to be checked
            % further by the derived class, usually in the (abstract)
            % method *init_system_parameters(...)* , which is normally called
            % at the beginning of the (abstract) method *solve(...)* . 
            % . 
            %
            %
            % Each of these parameters can take the following values v:
            %
            %   * v = -1, indicating that the parameter was not explicitly
            %     specified. Typically, a suitable default kicks in, which
            %     may depend on the derived BSolver class. 
            %   * 0<v<1, indicating that a fraction v of the number of
            %     available samples is used.
            %   * v = a positive integer, indicating that a fixed number of
            %     samples is used. If v exceeds the number of available
            %     samples, the number is clipped to the number of
            %     samples in the signal.
            %
            
            % create an empty Parameters object to store the solver
            % parameters
            self.parms = Parameters( class(self), {} );
        end
        %------------------------------------------------------------------------------------------
        b = solve( self, signal, varargin )
            % [abstract method] Estimate the normalized base parameters *b*
            % from *signal*. 
            %
            % :param signal: The :class:`Signal` object from which to
            %   estimate the base parameters. The signal is added as a
            %   required argument to the self.parms Parameters
            %   object.
            % :param varargin: any of the solver parameters in self.parms
            %   can be overwritten here.
            % :return: column vector with the estimated base parameters.
            %
        %------------------------------------------------------------------------------------------
        function init_system_parameters( self )
            % [virtual object method] % Derived classes may call this
            % method to validate the signal passed to the solve method, and
            % overwrite this method to initialize the parameters of the
            % system that has to be solved:
            %
            %   * self.gep.nsamples
            %
            % using the corresponding solver parameters in self.parms and 
            % the signal. 
            %
            % The nsamples parameter is always applied first. 

            if self.parms.p.signal.value.length==0
                Sparsimatics.Error( [class(self) ':init_system_parameters(...)'] ...
                  , 'Signal argument has no samples.' ...
                  );
            end

        end
        %--------------------------------------------------------------------------------------
        function cp = copy(self)
            % 
            % The standard copy performs a shallow copy::
            %
            %   copy_of_bsolver = my_bsolver.copy()
            %   assert(copy_of_my_bsolver~=my_bsolver) % this is ok
            %   assert(copy_of_my_bsolver.parms~=my_bsolver.parms) % oops, this fails
            % 
            % Because of the shallow copy the parms property of both
            % BSolvers is the same! Consequently, if we modify
            % _my_bsolver.parms_ , we also modify _copy_of_my_bsolver_ .
            %
            % This is obviously not what we want. This method performs a
            % deep copy.
            %
            
            % first make new (derived) BSolver object
            cp = feval(class(self));
            % make a shallow copy of the parms property
            new_parms = copy(self.parms);
            % and overwrite cp.parms
            cp.parms = new_parms;
        end
        %--------------------------------------------------------------------------------------
        function s = get_display_struct(self)
            s = struct;
            try 
                %self.parms.p.signal
                s.signal   = self.parms.p.signal.value.label;
            catch
            end
        end
        %--------------------------------------------------------------------------------------
    end
    %------------------------------------------------------------------------------------------
    methods (Access=protected)
        %--------------------------------------------------------------------------------------
        function propgrp = getPropertyGroups(self)
            if ~isscalar(self)
                propgrp = getPropertyGroups@matlab.mixin.CustomDisplay(self);
            else
                propgrp = matlab.mixin.util.PropertyGroup(self.get_display_struct);
            end
        end
        %------------------------------------------------------------------------------------------
    end    
    %------------------------------------------------------------------------------------------
end

