classdef BSolverGEP < BSolverSimple
    % A class for estimating (fitting) base terms from a *Signal*, using the Generalized Eigenvalue
    % problem (GEP) formulation, with square Hankel matrices.
    %
    % Inherits BSolver.
    %
    methods
        %------------------------------------------------------------------------------------------
        function self = BSolverGEP(varargin)
            % [constructor] Create a base term solver that uses the
            % Generalized Eigenvalue Problem parametric method. This solves
            % a system of square Hankel matrices of dimension nrows x ncols
            % with
            %
            %   * nterms = ncols = nrows,
            %   * nsamples_used = 2*nrows.
            %
            % These equalities imply that one can specify one parameter
            % only: nrows, ncols, nterms, or nsamples.
            %
            % :return: a BSolverGEP object.
            %
            % The nsamples parameter is always applied first. 
            % E.G. if a signal has 100 samples, and the nsamples parameter
            % is set to .9 only 90 samples are used. If, additionally, the
            % nrows parameter is set to .50, a system of 0.50x90 = 45 rows
            % is constructed. 
            %
            % `examples <tutorial_BSolver.html>`_
            %
            
            self.init( varargin{:} );
        end
        %------------------------------------------------------------------------------------------
        function init_system_parameters( self )
            % [object method] initialize solver parameters.
            % GEP uses two square Hankel matrices of size mxm and 2m
            % samples.
       
            % only one parameter must be specified
            n_specified = (self.parms.p.nsamples.value~=-1) ...
                         +(self.parms.p.nrows.value   ~=-1) ...
                         +(self.parms.p.ncols.value   ~=-1) ...
                         +(self.parms.p.nterms.value  ~=-1);
            if n_specified > 1
                % too much information
                Sparsimatics.Error( 'BSolverGEP.init_system_parameters'...
                  ,[    'Only one of the parameters ''--nrows'', ''--ncols'', ''--nterms'','...
                    '\n! ''--nsamples'' may be specified, got %d.']...
                  , n_specified ...
                  );
            end
            
            self.init_system_parameters@BSolverSimple();
            nsamples = self.gep.nsamples;
            
            if self.parms.p.nsamples.value~=-1 || n_specified == 0
                % system size (hankel matrix)
                % GEP needs even number of samples, discard one sample if odd:
                m = floor(nsamples/2);                
            else
                if self.parms.p.nrows.value~=-1
                    m_param = self.parms.p.nrows.value;
                elseif self.parms.p.ncols.value~=-1
                    m_param = self.parms.p.ncols.value;
                elseif self.parms.p.nterms.value~=-1
                    m_param = self.parms.p.nterms.value;
                end
                m = apply_fraction_or_min(m_param,floor(nsamples/2));
            end
            
            self.gep.nsamples = 2*m;
            self.gep.nrows    = m;
            self.gep.ncols    = m;
            self.gep.nterms   = m;

            Sparsimatics.Warnv(2,'! Info','BSolverGEP.init_system_parameters'...
              , ['Solving %dx%d generalized eigenvalue problem for %d terms, using %d samples.'...
                ,'\n! Algorithm used is Generalized Eigenvalue Problem (square).'....
                ]...
              , self.gep.nrows, self.gep.ncols, self.gep.nterms, self.gep.nsamples ...
              );
        end
        %------------------------------------------------------------------------------------------
        function b = solve( self, varargin )
            % [object method] Estimate the normalized base parameters *b* from *signal*. 
            % Infinite coefficients are removed. 
            % Set *Sparsimatics.Verbosity>0* to report guesses.
            %
            % :param key,value,double [signal]: The *Signal* object from which to estimate the base parameters.
            % :return: column vector with the estimated base parameters.
            %
            % `examples <tutorial_BSolver.html>`_
            %
            self.parms.add_parameters( {'signal:Signal',Required}, varargin{:} );
            
            self.init_system_parameters()
            
            m = self.gep.nrows;
            H0 = hankel( self.parms.p.signal.value.samples(1:m  ), self.parms.p.signal.value.samples(m  :2*m-1) );
            H1 = hankel( self.parms.p.signal.value.samples(2:m+1), self.parms.p.signal.value.samples(m+1:2*m  ) );
            
            size0 = size(H0);
            if ~(   size0(1)==size0(2) ) ...
            || ~( all(size0==size(H1)) )
                Sparsimatics.Error( fa, 'BSolverGEP: implementation error');
            end
            b = eig(H1,H0);

            k = find( isinf(b) );
            if ~isempty(k)
                Sparsimatics.Warn( self.parms.get_signature ...
                    ,[      'Some base terms (#=%d) have been discarded because they are infinite.' ...
                      '\n    Discarded terms: %d.\n'] ...
                    , length(k), k ...
                    );
                b(k) = [];
            end
            b = b.';
        end
        %------------------------------------------------------------------------------------------
    end    
end

