classdef BSolverSimple < BSolver
    % Abstract class for estimating (fitting) base terms from a
    % :class:`Signal`. Derived classes, generally referred to as
    % *BSolvers*, must implement a:: 
    %
    %   b = BSolverSubclass.solve( self, signal )
    %
    % method that computes the normalized base terms from a :class:`Signal`
    % object. All BSolvers inherit class *BSolver*.
    %
    properties(SetAccess=protected,GetAccess=public)
        gep = struct('nsamples', -1 ...
                    ,'nrows'   , -1 ...
                    ,'ncols'   , -1 ...
                    ,'nterms'  , -1 ...
                    );
    end
     
    methods
        %------------------------------------------------------------------------------------------
        function init( self, varargin )
            % [delayed constructor] Initialize (or re-initialize) a
            % BSolver. This sets the property self.parms, which is a
            % Parameters object keeping track of the solver's
            % parameters:
            %
            %   * self.parms.p.nrows,
            %   * self.parms.p.ncols,
            %   * self.parms.p.nsamples, and
            %   * self.parms.p.nterms.
            %
            % These properties can also be set directly on a given BSolver
            % object, e.g.::
            %
            %   my_bsolver.parms.p.nrows.set_value(5)
            %
            % The init method checks that these properties are in the set
            % { -1, x:0<x<1, 1, 2, 3, ...}. The validity of the parameter
            % combination is to be validated further by the derived class,
            % usually in the init_system_parameters (abstract) method,
            % which is normally called at the beginning of the solve
            % (abstract) method. 
            %
            % :param key,value=double [nsamples]: upper limit on the number
            %   of samples in the input signal that the solver can use
            %   (-1, in ]0,1[, or integer>1). This parameter will always be
            %   applied before the other solver parameters.
            % :param key,value=double [nrows]: the number of system rows
            %   that the solver will use (integer or in ]0,1[). 
            % :param key,value=double [ncols]: the number of system
            %   columns that the solver will use (integer or in ]0,1[). 
            % :param key,value=double [nterms]: the number of base terms
            %   that the solver will solve for(integer or in ]0,1[). 
            %
            % Each of these parameters can take the following values v:
            %
            %   * v = -1, indicating that the parameter was not explicitly
            %     specified. Typically, a suitable default kicks in, which
            %     may depend on the derived BSolver class. 
            %   * 0<v<1, indicating that a fraction v of the number of
            %     available samples is used.
            %   * v = a positive integer, indicating that a fixed number of
            %     samples is used. If v exceeds the number of available
            %     samples, the number is clipped to the number of
            %     samples in the signal.
            %
            % :return: self.
            %
            % Example::
            %
            %   bsolver.init('--nterms', 20, '--nrows', 0.60);
            %
            % Alternatively::
            %
            %   bsolver.parms.p.nterms.set_value(20)   % solve for 20 terms
            %   bsolver.parms.p.nrows.set_value(0.60)  % use 0.60*nsamples rows
            %
            self.init@BSolver();
            
            self.parms.add_parameters ...
               ( { 'nsamples:@is_minus1_or_fraction_or_positive_integer', -1 ...
                 , 'nrows   :@is_minus1_or_fraction_or_positive_integer', -1 ...
                 , 'ncols   :@is_minus1_or_fraction_or_positive_integer', -1 ...
                 , 'nterms  :@is_minus1_or_fraction_or_positive_integer', -1 ...
                 } ...
               , varargin{:} ...
               );
        end
        %------------------------------------------------------------------------------------------
        b = solve( self, signal, varargin )
            % [abstract method] Estimate the normalized base parameters *b*
            % from *signal*. 
            %
            % :param signal: The :class:`Signal` object from which to
            %   estimate the base parameters. The signal is added as a
            %   required argument to the self.parms Parameters
            %   object.
            % :param varargin: any of the solver parameters in self.parms
            %   can be overwritten here.
            % :return: column vector with the estimated base parameters.
            %
        %------------------------------------------------------------------------------------------
        function init_system_parameters( self )
            % [virtual object method] % Derived classes may call this
            % method to validate the signal passed to the solve method, and
            % overwrite this method to initialize the parameters of the
            % system that has to be solved:
            %
            %   * self.gep.nsamples
            %   * self.gep.nrows
            %   * self.gep.ncols 
            %   * self.gep.nterms
            %
            % using the corresponding solver parameters in self.parms and 
            % the signal. 
            %
            % The nsamples parameter is always applied first. 
            %
            % Here are some examples, assuming a signal containing 100
            % samples. Unspecified parameters are -1.
            %
            %   1. nsamples=.90: only 90 (=.90x100) samples are used.
            %   2. nrows=.50: a system with 50 (=.50x100) rows is
            %      constructed.
            %   3. nsamples=.90, nrows=.50: only 90 (=.90x100) samples are
            %      used, and a system with 45 (=.50x90) rows is
            %      constructed.
            %   4. nsamples=80: only 80 samples are used.
            %   5. nrows=50: a system with 50 rows is constructed.
            %   6. nsamples=120: request using 120 samples, but since the
            %      signal contains only 100 samples, the value is clipped
            %      to 100.
            %
            self.init_system_parameters@BSolver()
            
            self.gep.nsamples = ...
                apply_fraction_or_min( self.parms.p.nsamples.value ...
                                     , self.parms.p.signal.value.length ...
                                     );
        end
        %--------------------------------------------------------------------------------------
        function cp = copy(self)
            % 
            % The standard copy performs a shallow copy::
            %
            %   copy_of_bsolver = my_bsolver.copy()
            %   assert(copy_of_my_bsolver~=my_bsolver) % this is ok
            %   assert(copy_of_my_bsolver.parms~=my_bsolver.parms) % oops, this fails
            % 
            % Because of the shallow copy the parms property of both
            % BSolvers is the same! Consequently, if we modify
            % _my_bsolver.parms_ , we also modify _copy_of_my_bsolver_ .
            %
            % This is obviously not what we want. This method performs a
            % deep copy.
            %
            
            % first make new (derived) BSolver object
            cp = feval(class(self));
            % make a shallow copy of the parms property
            new_parms = copy(self.parms);
            % and overwrite cp.parms
            cp.parms = new_parms;
        end
        %--------------------------------------------------------------------------------------
        function s = get_display_struct(self)
            s = get_display_struct@BSolver(self);
            try 
                self.parms.p.signal
%                 s.signal   = self.parms.p.signal.value.label;
                s.nsamples = [ num2str(self.parms.p.nsamples.value) ' -> ' num2str(self.gep.nsamples ) ];
                s.nrows    = [ num2str(self.parms.p.nrows.value)    ' -> ' num2str(self.gep.nrows    ) ];
                s.ncols    = [ num2str(self.parms.p.ncols.value)    ' -> ' num2str(self.gep.ncols    ) ];
                s.nterms   = [ num2str(self.parms.p.nterms.value)   ' -> ' num2str(self.gep.nterms   ) ];
            catch
                s.nsamples = self.parms.p.nsamples.value;
                s.nrows    = self.parms.p.nrows.value;
                s.ncols    = self.parms.p.ncols.value;
                s.nterms   = self.parms.p.nterms.value;
            end
        end
        %--------------------------------------------------------------------------------------
    end
    %------------------------------------------------------------------------------------------
    methods (Access=protected)
        %--------------------------------------------------------------------------------------
        function propgrp = getPropertyGroups(self)
            if ~isscalar(self)
                propgrp = getPropertyGroups@matlab.mixin.CustomDisplay(self);
            else
                propgrp = matlab.mixin.util.PropertyGroup(self.get_display_struct);
            end
        end
        %------------------------------------------------------------------------------------------
    end    
    %------------------------------------------------------------------------------------------
end

