classdef CSolverVandermondeLS < CSolver
    % A CSolver that estimates the *c* coefficients by solving a least squares Vandermonde
    % system.
    %
    % Inherits CSolver.
    %
    
    methods
        %------------------------------------------------------------------------------------------
        function self = CSolverVandermondeLS(varargin)
            % [Constructor] 
            self.init(varargin{:});
        end
        %------------------------------------------------------------------------------------------
        function init_system_parameters(self)
            % [object method] Initialize solver parameters, relative to
            % signal and b.
                        
            if self.parms.p.nsamples.value~=-1
                self.vdm.nsamples = apply_fraction_or_min(self.parms.p.nsamples.value,self.parms.p.signal.value.length);
            else
                self.vdm.nsamples = self.parms.p.signal.value.length;
            end
            
            nb = length(self.parms.p.b.value);
            self.vdm.ncols =nb;
            if self.parms.p.delta.value==-1
                if self.parms.p.nrows.value==-1
                    self.vdm.nrows = self.vdm.nsamples;
                    self.vdm.delta = 1;
                else
                    self.vdm.nrows = apply_fraction_or_min(self.parms.p.nrows.value,self.vdm.nsamples);
                    nb = length(self.parms.p.b.value);
                    if self.vdm.nrows<nb
                        Sparsimatics.Error('CSolverVandermondeLS.init_system_parameters()'...
                          ,['Requesting less rows (%d) than there are exponential terms (%d) yields '...
                            'an underdetermined system. Increase the --nrows option to at least %d.']...
                          ,self.vdm.nrows, nb, nb ...
                          );
                    end
                    if self.vdm.nrows==1
                        self.vdm.delta = 1;
                    else
                        self.vdm.delta = floor((self.vdm.nsamples-1)/(self.vdm.nrows-1));
                    end
                end
            else
                self.vdm.delta = self.parms.p.delta.value;
                if self.parms.p.nrows.value==-1
                    %use as many rows as possible
                    self.vdm.nrows = 1 + floor((self.vdm.nsamples-1)/self.vdm.delta);
                else
                    self.vdm.nrows = self.parms.p.nrows.value;
                    nsamples_needed = 1 + (self.vdm.nrows-1)*self.vdm.delta;
                    if nsamples_needed>self.vdm.nsamples
                        Sparsimatics.Error( 'CSolverVandermondeLS.init_system_parameters()' ...
                          , ['Signal does not have enough samples to construct %d rows spaced at every %d-th sample. '...
                             'Allow for more samples, less rows (--nrows) or a smaller spacing (--delta), needed:%d, have:%d']...
                          , self.vdm.nrows, self.vdm.delta, nsamples_needed, self.vdm.nsamples ...
                          );
                    end
                end
            end
            
            Sparsimatics.Warnv(2,'! Info','CSolverVandermondeLS.init_system_parameters'...
              , ['Solving %dx%d Vandermonde system for %d coefficients, using %d samples, delta=%d.'...
                ,'\n! Algorithm used is Least Squares.'....
                ]...
              , self.vdm.nrows, self.vdm.ncols, self.vdm.ncols, self.vdm.nsamples/self.vdm.delta, self.vdm.delta ...
              );
        end
        %------------------------------------------------------------------------------------------
        function c = solve( self, varargin )
            % [object method] Estimate the *c* coefficients from *signal* for given base
            % parameters *b*. 
            %
            % :param signal: Signal object from which to estimate the *c*
            %   coefficients. 
            % :param b: array of base parameters for which the *c*
            %   coefficients will be estimated. 
            % :param remove: array of integers (>0) indicating rows that
            %   have to be removed from the vandermonde matrixand the right
            %   hand side. 
            % :return: array of *c* coefficients (row vector).
            %
            self.parms.add_parameters ...
              ( { 'signal:Signal:input signal'                          , Required ...
                , 'b:double:base parameters'                            , Required ...
                ,['remove_rows:@isempty|@is_integer&@is_strictly_positive:' ...
                  'rows to be removed from vandermonde matrix and rhs'] , [] ...
                } ...
              , varargin{:}...
              );
            
            self.init_system_parameters()
            
%             tic;
            
            vdm = vandermonde_jk( self.parms.p.b.value, self.vdm.nrows, 0, self.vdm.delta );
            rhs = transpose( self.parms.p.signal.value.samples( 1 : self.vdm.delta : self.vdm.delta*self.vdm.nrows));
            
            % remove rows in self.parms.p.remove.value from vandermonde
            % matrix and rhs:
            if ~isempty(self.parms.p.remove_rows.value)
                rows = true(numel(rhs),1);
                rows( self.parms.p.remove_rows.value ) = false;
                vdm = vdm(rows,:);
                rhs = rhs(rows);
            end
            % solve
            
            c = transpose( vdm \ rhs );
            
%             t=toc; fprintf('CSolverVdmLS  solve (w/o init): %f\n',t);
        end
        %------------------------------------------------------------------------------------------
    end    
end

