classdef MultiExponential < handle %& matlab.mixin.CustomDisplay
    % Class *MultiExponential* exposes methods for model parameter conversion. It is mainly used
    % internally in the *MultiExponentialParameters* class, but can occasionaly be useful by itself.
    %
    % Technically, this is a Matlab *class*, but since it has only *Static* methods, it acts as a
    % *namespace*, and its static methods as functions in that namespace.
    %
    methods(Static)
        %------------------------------------------------------------------------------------------            
        function tf = Is_angular_frequency_used(frequency_unit)
            %[class method] does the string *frequency_unit* correspond to
            % angular frequency (radians/s) or not (Hz)?
            if startsWith('radians/s',frequency_unit)
                tf = true;
            elseif startsWith('Hz',frequency_unit)
                tf = false;
            else
                Sparsimatics.Error( ['MultiExponential.Is_angular_frequency_used(' frequency_unit ')'] ...
                    , 'argument must be ''r[adians/s]'' or ''H[z]''.' ...
                    );
            end
        end
        %------------------------------------------------------------------------------------------            
        function [a,d,f,p] = Convert_bc_to_adfp( varargin )
            %[class method] Convert PHYSICAL base terms and coefficients
            % to physical quantities:  amplitudes, damping factors,
            % frequencies, and phases.  
            %
            % :param b: array of base terms.
            % :param c: array of coefficients.  
            % :param key,value=char frequency_unit: return frequency in
            %   radians/s (i.e. the angular frequency) or Hz, either
            %   'r[adians/s]'|'H[z]'. Default is angular frequency units
            %   ('r'). 
            % :returns: *[a,d,f,p]* arrays of amplitudes [arbitrary units],
            %   damping factors [Hz], frequencies [Hz|radians/s], and
            %   phases [radians]. 
            fa = Parameters ...
                   ( ['MultiExponential.Convert_bc_to_adfp:Convert {base terms,coefficients} ' ...
                      'to {amplitudes,damping factors,frequencies,phases}.'] ...
                   , {'b:double'                                                      , Required ...
                     ,'c:double'                                                      , Required ...
                     ,'frequency_unit:char:frequency unit ''r[adians/s]'' or ''H[z]''', 'r' ...
                     } ...
                   , varargin{:} ...
                   );
            b = fa.p.b.value;
            c = fa.p.c.value;

            n = length(b);
            if ~( n == length(c) )
                Sparsimatics.Error( fa, ': b and c must have the same length: %d~=%d.', n, length(c) );
            end
            
%           this is a waste of time:
%             b = reshape(b,[1,n]);
%             c = reshape(c,[1,n]);
            
            a = abs(c); %complex magnitude
            p = angle(c);
            
            logb = log(b);
            
            d = real(logb);
            if MultiExponential.Is_angular_frequency_used(fa.p.frequency_unit.value)
                f = imag(logb);
            else
                f = imag(logb)/(2*pi);
            end
        end
        %------------------------------------------------------------------------------------------            
        function [b,c] = Convert_adfp_to_bc( varargin )
            %[class method] Convert the physical quantities *amplitudes*,
            % *damping* *factors*, *frequencies* and *phases* to physical
            % or normalized base terms and coefficients. 
            %
            % :param amplitudes: real array of amplitudes [arbitrary units].
            % :param damping_factors: real array of damping factors [1/s].
            %   A positive value indicates exponential growth, a negative
            %   number indicates exponential decay (=damping)
            % :param frequencies: real array of frequencies.
            % :param phases: the array of phases [radians].
            % :param key,value=double [sampling_frequency]: if a sampling
            %   frequency is provided, the base parameters are normalized
            %   by that sampling frequency.  
            % :param char frequency_unit: used frequency unit: radians/s
            %   (i.e. the angular frequency) or Hz 'r[adians/s]'|'H[z]'.
            %   Default is angular frequency units.
            % :return: *[b,c]*, array of physical or normalized base terms
            %   and corresponding coefficients. 
            %
            % All input arrays must have the same length. Output arrays
            % have the same length. 
            %
            fa = Parameters...
                   (['MultiExponential.Convert_adfp_to_bc:Convert {amplitudes,damping' ...
                     ' factors,frequencies,phases} to {base terms,coefficients}.']...
                   , {'amplitudes        :@is_strictly_positive: [arbitrary units]'   , Required ...
                     ,'damping_factors   :double: [1/s] decay if <0, growth if >0 '   , Required ...
                     ,'frequencies       :double: frequency or angular frequency'     , Required ...
                     ,'phases            :double               : [radians]'           , Required ...
                     ,'sampling_frequency:@is_strictly_positive: [Hz]'                , 1 ...
                     ,'frequency_unit:char:frequency unit ''r[adians/s]'' or ''H[z]''', 'r' ...
                     } ...
                   , varargin{:} ...
                   );
            a = fa.p.amplitudes.value;
            d = fa.p.damping_factors.value;
            f = fa.p.frequencies.value;
            p = fa.p.phases.value;
            sampling_frequency = fa.p.sampling_frequency.value;
                       
            nterms = length(a);
            if length(d)~=nterms ...
            || length(f)~=nterms ...
            || length(p)~=nterms 
                Sparsimatics.Error( fa, 'a, d, f and p must be of equal length.');
            else
                a = reshape( a, [1,nterms] );
                d = reshape( d, [1,nterms] );
                f = reshape( f, [1,nterms] );
                p = reshape( p, [1,nterms] );
            end
            
            c = a .* exp( 1i*p );
        
            if MultiExponential.Is_angular_frequency_used(fa.p.frequency_unit.value)
                i = 1i;
            else
                i = 2*pi*1i;
            end
            b = exp( ( d + f*i )/sampling_frequency );
        end
        %------------------------------------------------------------------------------------------            
        function b = Convert_beta_to_b( varargin )
            % [class method] Convert the normalized base terms to physical base terms:
            %
            % :param beta: array with normalized base parameters.
            % :param sampling_frequency: sampling frequency [Hz] with which the *beta* are normalized.
            % :return: array with corresponding physical base parameters.
            %
            fa = Parameters...
                   ( ['MultiExponential.Convert_beta_to_b:Convert ' ...
                      'normalized base terms to physical base terms.'] ...
                   , {'beta:double: normalized base terms'              , Required ...
                     ,'sampling_frequency:@is_strictly_positive: [Hz]'  , Required ...
                     } ...
                   , varargin{:} ...
                   );
            beta = fa.p.beta.value;
            sampling_frequency = fa.p.sampling_frequency.value;

            if sampling_frequency==1
                b = beta;
            else
                b = exp( log(beta)*sampling_frequency );
            end
            if size(b,1)~=1
                b = reshape(b,[1,length(b)]);
            end
        end
        %------------------------------------------------------------------------------------------            
        function beta = Convert_b_to_beta( varargin )
            % [class method] Normalize the physical base parameters by a sampling frequency. 
            %
            % :param b: array with physical base parameters.
            % :param sampling_frequency: sampling frequency [Hz] with which the base parameters will be normalized.
            % :return: array with the normalized base parameters.
            %
            fa = Parameters ...
                   ( ['MultiExponential.Convert_b_to_beta' ...
                     ,':Convert physical base terms to normalized base terms.'] ...
                   , {'b:double: normalized base terms', Required ...
                     ,'sampling_frequency:double: [Hz]', Required ...
                     } ...
                   , varargin{:} ...
                   );
            b                  = fa.p.b.value;
            sampling_frequency = fa.p.sampling_frequency.value;
            
            if sampling_frequency==1
                beta = b;
            else
                beta = exp( log(b)/sampling_frequency );
            end
            if size(beta,1)~=1
                beta = reshape(beta,[1,length(beta)]);
            end
        end
        %------------------------------------------------------------------------------------------            
        function new_beta = Convert_beta_to_beta( varargin)
            % [class method] Convert the normalized base terms *old_beta* from
            % *old_sampling_frequency* to *new_sampling_frequency*. This conversion may suffer from
            % numerical accuracy problems.
            %
            % :param old_beta: array with 'old' normalized base parameters
            % :param old_sampling_frequency: sampling frequency [Hz] with which *old_beta* was normalized.
            % :param new_sampling_frequency: sampling frequency [Hz] with which *new_beta* will be normalized [Hz].
            % :return: array of 'new' base parameters *new_beta* , normalized with *new_sampling_frequency*.
            %
            fa = Parameters...
                   ( 'MultiExponential.Convert_beta_to_beta: Convert normalized base terms from one sampling_frequency to another.' ...
                   , {'old_b:double: normalized base terms'                          , Required ...
                     ,'old_sampling_frequency:@is_strictly_positive: [Hz]'           , Required ...
                     ,'new_sampling_frequency:@is_strictly_positive: [Hz]'           , Required ...
                     } ...
                   , varargin{:} ...
                   );
            new_beta = exp( log(fa.p.old_b.value)*fa.p.old_sampling_frequency.value/fa.p.new_sampling_frequency.value );
        end
        %------------------------------------------------------------------------------------------            
    end
    %------------------------------------------------------------------------------------------            
end

