classdef MultiExponentialParameters < handle & matlab.mixin.Copyable
    % This class is a simple data class for storing the parameters of a multi-exponential:
    %
    %   .. math::
    %       f(t) = \sum_{i=1}^n c_i b_i^t, 
    % 
    % where :math:`t` represents time, and :math:`c_i` and :math:`b_i` are complex numbers. The
    % :math:`b_i` are referred to as the *base* *terms* or *base* *parameters*, and the :math:`c_i`
    % as the *coefficients*. For an equidistantly sampled signal with sampling frequency :math:`F_s`
    % and sampling interval :math:`\Delta=1/F_s` 
    % 
    %   .. math::
    %       t = 0, 1/F_s, 2/F_s, ... =  0, \Delta, 2\Delta, ...  
    % 
    % The :math:`b_i` and the time :math:`t` are often normalised by the sampling frequency :math:`F_s`: 
    %
    %   .. math::
    %       f(t) = \sum_{i=1}^n c_i \beta_i^\tau, 
    %
    % with :math:`\beta_i = \mathrm{e}^{ \frac{\log(b_i)}{F_s} }` the *normalized* base terms, and :math:`\tau =
    % t F_s`, the *normalized* time, which now takes discrete integer values:
    %
    %   .. math::
    %       \tau = 0, 1, 2, ..., 
    %1:k:k*nb
    % Generally, the normalized base terms are close to the unit circle in the complex plane and
    % have they have many numerical advantages over their physical counterparts. In Sparsimatics, if
    % the nature of the base terms is unspecified, the normalized base terms are always understood. 
    %
    % For examples on using this class see `tutorial_MultiExponential`_.
    %
    properties (SetAccess=private)
        sampling_frequency=0 % [read-only property] The sampling_frequency used for normalization.
        b=[]                 % [read-only property] The normalized base terms in the exponential expansion.
        c=[]                 % [read-only property] The coefficients in the exponential expansion.
    end
    methods
        %------------------------------------------------------------------------------------------            
        function self = MultiExponentialParameters( varargin )
            % [constructor] This method initializes the normalized base terms and the coefficients
            % of the exponential expansion, and the sampling frequency used in the normalization.
            %
            % :param sampling_frequency: the sampling frequency unsed for normalizing the base terms.
            % :param parameters: a cell array containing the parameters from which the object is constructed (see below).
            % :param time_scale: a case insensitive string, either 'p[hysical]' or 'n[ormalized]', indicates whether the parameters correspond to a physical or normalized time scale.
            % :return: a *MultiExponentialParameters* object.
            %
            % The *parameters* argument contains, 
            %
            %   * either the base terms *b* (physical or normalized,according to the *time_scale*
            %     parameter) and the coefficients *c*, 
            %   * or the amplitudes *a*, the damping factors *d*, the frequencies *f* and the phases
            %     *p*. The latter are physical quantities expressed in the following units, and commonly denoted by the following symbols: 
            %
            %       * *a*, amplitudes [arbitrary units, the same as the signal unit], $\beta_i$ (not to be mistaken for the normalized base parameters).
            %       * *d*, damping factors [1/s] $\psi_i$.
            %       * *f*, frequencies [Hz = cycles/s], $\omega_i$.
            %       * *p*, phases [radians], $\gamma_i$.
            %
            %     If these are used with 'normalized' as the *time_scale* parameter, they are
            %     interpreted as {*a*, *d/Fs*, *f/Fs*, *p*}, *Fs* being the sampling frequency. 
            %
            % `examples <examples_MultiExponentialParameters_ctor.html>`_
            %
            self.init( varargin{:} );
        end
        %------------------------------------------------------------------------------------------            
        function self = init( self, varargin )
            % [object method] A delayed constructor with the same signature as the
            % constructor :meth:`MultiExponentialParameters.MultiExponentialParameters`. 
            %
            % `MultiExponentialParameters() examples <examples_MultiExponentialParameters_ctor.html>`_
            %
            fa = Parameters( 'MultiExponentialParameters.init:delayed constructor' ...
                   , {'sampling_frequency:@is_strictly_positive: [Hz]'                 , Required ...
                     ,'parameters:@is_multi_exponential_parameters: {b,c} or {a,d,f,p}', Required ...
                     ,'time_scale:char: ''normalized'' or ''physical'''                , Required ...
                     ,'frequency_unit:char:frequency unit ''r[adians/s]'' or ''H[z]''' , 'r'      ...
                     } ...
                   , varargin{:} ...
                   );
               
            self.sampling_frequency = fa.p.sampling_frequency.value;
            %assert_is_strictly_positive( self.sampling_frequency, 'sampling_frequency', fa );            

            physical = is_physical(fa.p.time_scale.value);

            switch length(fa.p.parameters.value)
                case 2 % {b,c}
                    [self.b,self.c] = fa.p.parameters.value{:};
                    
                    r = row_vector(self.b);
                    if ~isempty(r)
                        self.b = r;
                    end
                    
                    r = row_vector(self.c);
                    if ~isempty(r)
                        self.c = r;
                    end
                    
                    if physical
                        self.b = MultiExponential.Convert_b_to_beta( self.b, self.sampling_frequency );
                    end
                    
                case 4 % {a,d,f,p}
                    [a,d,f,p] = fa.p.parameters.value{:};
                    
                    r = row_vector(a);
                    if ~isempty(r)
                        a = r;
                    end
                    
                    r = row_vector(d);
                    if ~isempty(r)
                        d = r;
                    end
                    
                    r = row_vector(f);
                    if ~isempty(r)
                        f = r;
                    end
                    
                    r = row_vector(p);
                    if ~isempty(r)
                        p = r;
                    end
                    
                    if ~physical
                        d = d*self.sampling_frequency;
                        f = f*self.sampling_frequency;
                    end
                    [self.b,self.c] = MultiExponential.Convert_adfp_to_bc  ...
                       ( a, d, f, p                                        ...
                       , '--sampling-frequency', self.sampling_frequency   ...
                       , '--frequency-unit'    , fa.p.frequency_unit.value ...
                       );
                    
            end
        end
        %------------------------------------------------------------------------------------------            
        function b_phys = physical_b(self)
            % [object method] Retrieve this object's physical base terms.
            %
            % :return: *b*, the **physical** base terms.
            %
            % `MultiExponentialParameters,physical_parameters() examples <examples_MultiExponentialParameters_ctor.html>`_
            %
            if self.sampling_frequency==1
                b_phys = self.b;
            else
                b_phys = exp( log(self.b)*self.sampling_frequency );
            end
        end
        %------------------------------------------------------------------------------------------            
        function a = amplitude(self)
            % :return: the amplitudes of the respective terms.
            a = abs(self.c);
        end
        %------------------------------------------------------------------------------------------            
        function d = damping(self)
            % :return: the damping factors of the respective terms. If <0
            %   damping, if >0 exponential growth.
            d = real(log(self.b));
        end
        %------------------------------------------------------------------------------------------
        % 1 Hz = 2*pi radians/s
        % f = 2*pi*w (w = omega = Hz = frequency, f = angular frequency)
        function [f,d] = frequency(self)
            % :return: the frequencies of the respective terms [radians/s]
            %   and optionally, also the damping factors. If d<0 the term
            %   is damped, if d>0 it is exponentially growing.
            logb = log(self.b);
            f = imag(logb) * self.sampling_frequency;
            if nargout>1
                d = real(logb);
            end
        end
        %------------------------------------------------------------------------------------------            
        function [w,d] = frequency_Hz(self)
            % :return: the frequencies of the respective terms [Hz].
            %   and optionally, also the damping factors. If d<0 the term
            %   is damped, if d>0 it is exponentially growing.
            logb = log(self.b);
            w = imag(logb) * self.sampling_frequency / (2*pi);
            if nargout>1
                d = real(logb);
            end
        end
        %------------------------------------------------------------------------------------------            
        function p = phase(self)
            % :return: the phases of the respective terms.
            p = angle(self.c);
        end
        %------------------------------------------------------------------------------------------            
        function rv = nterms(self)
            % [object method] Return the number of terms in the multi-exponential.
            %
            % :return: the number of terms in the multi-exponential
            %
            % `MultiExponentialParameters.nterms() examples <examples_MultiExponentialParameters_energy.html>`_
            %
            rv = length(self.c);
        end
        %------------------------------------------------------------------------------------------            
        function E = energies(self,varargin)
            % [object method] Compute and return the energy of each
            % multi-exponential term. It is defined as:
            %
            %   .. math::
            %       E_i = \sum_{j=1}^m |c_i b_i^{t_j}|^2, i=1..n
            %
            % with :math:`n` the number of terms in the multi-exponential
            % model, and :math:`t_j, j=1..m` the time points in the sample. 
            %
            % :param nsamples: a number of samples or a signal from which
            %   the number of samples and t0 are obtained. 
            % :param key,value=double [downsampling_factor]: undersampling
            %   factor. Default = 1.  
            % :param key,value=double [t0]: initial time point. Default is
            %   0s, or signal.t0 if the first argument is a *Signal*. 
            % :return: row vector with the energies of each term.
            %
            % If *nsamples* is a Signal with a different sampling_frequency than the base terms are
            % rescaled by the sampling_frequency of the _Signal_.
            %
            % `MultiExponentialParameters.energies() examples <examples_MultiExponentialParameters_energy.html>`_
            %
            trms = self.terms( varargin{:} );
            E = dot(trms,trms,1);
        end
        %------------------------------------------------------------------------------------------            
        function trms = terms(self,varargin)
            % [object method] Compute and return the individual terms of
            % each multi-exponential term. It is defined as:
            %
            %   .. math::
            %       T_i = \sum_{j=1}^m |c_i b_i^{t_j}|^2, i=1..n
            %
            % with :math:`n` the number of terms in the multi-exponential
            % model, and :math:`t_j, j=1..m` the time points in the sample. 
            %
            % :param nsamples: a number of samples or a signal from which
            %   the number of samples and t0 are obtained. 
            % :param key,value=double [downsampling_factor]: undersampling
            %   factor. Default = 1. 
            % :param key,value=double [t0]: initial time point. Default is
            %   0s, or signal.t0 if the first argument is a *Signal*. 
            % :return: matrix with one row per term and one column per time
            %   point. 
            %
            % If *nsamples* is a Signal with a different sampling frequency
            % than self the base terms are rescaled by the sampling
            % frequency of the _Signal_. 
            %
            % This function has the same signature as
            % _MultiExponentialParameters.energies_. 
            %
            % `MultiExponentialParameters.terms() examples
            % <examples_MultiExponentialParameters_energy.html>`_ . 
            %
            fa = Parameters...
                   ( ['MultiExponentialParameters.terms:Compute the contribut'  ...
                     ,'ion of the individual components to a synthetic signal.']...
                   , {'nsamples:@is_integer|Signal: number of samples'       , Required ...
                     ,'downsampling_factor:@is_integer: undersampling factor', 1 ...
                     ,'t0:double: first time point [s]'                      , 0 ...
                     } ...
                   , varargin{:} ...
                   );
            nsamples            = fa.p.nsamples.value;
            downsampling_factor = fa.p.downsampling_factor.value;
            t0                  = fa.p.t0.value;
            
            b_ = self.b ;
            Fs = self.sampling_frequency;
            if isa(nsamples,'Signal')
                signal = nsamples;
                if t0~=signal.t0  && ~fa.was_provided('--t0')
                    t0 = signal.t0;
                    Sparsimatics.Warn(fa, 'The terms are computed using the signal''s t0 (%ds)',t0);
                end
                nsamples = signal.length();
                if nsamples>0
                    if Fs~=signal.sampling_frequency
                        Sparsimatics.Warn( fa, ['The terms were computed after converting the base terms to the signal''s sampling_frequency:\n' ...
                                                '! (%dHz->%dHz)'], self.sampling_frequency, signal.sampling_frequency );  
                        b_ = MultiExponential.Convert_beta_to_beta( b_, Fs, signal.sampling_frequency );
                        Fs = signal.sampling_frequency;
                    end
                end
             end
                        
            n_terms = self.nterms;
            if ~nsamples>0
                Sparsimatics.Warning( fa, 'Number of samples is 0.');
                trms = [];
            else
                if ~( n_terms > 0 )
                    Sparsimatics.Error( fa, 'no base terms found.');
                end
                trms = zeros(nsamples,n_terms);
                trms(1,:) = self.c .* b_.^(t0*Fs); % t0 is physical time!
                if downsampling_factor~=1
                    b_ = b_ .^ downsampling_factor;
                end
                for it=1+downsampling_factor:downsampling_factor:nsamples
                    trms(it,:) = trms(it-downsampling_factor,:) .* b_ ;
                end
            end            
        end
        %------------------------------------------------------------------------------------------            
        function nb = remove(self,k)
            % [object method] Remove terms with index k from the expansion.
            %
            % :param k: array of indices to be removed.
            % :return: number of terms left
            %
            % `MultiExponentialParameters.remove() examples <examples_MultiExponentialSolver.html>`_
            %
            % `filtering examples <examples_MultiExponentialSolver_filtering.html>`_
            %
            self.b(k) = [];
            self.c(k) = [];
            nb = length(self.b);
            if nb==0
                Sparsimatics.WARN('Warning', 'MultiExponentialParameters.remove(k)', 'Removed all terms. Object is empty.');
            end
        end
        %------------------------------------------------------------------------------------------                    
        function signal_values = compute_signal_at( self, varargin )
            % [class method] Compute the values of the synthetic signal at
            % arbitrarty physical or normalized time points [s].
            %
            % :param t: number of samples in the constructed signal
            %   (a positive integer). 
            % :param key,value=char [t0]: 'n[ormalized]' (default) or'p[hysical]'
            
            fa = Parameters ...
                  ( ['MultiExponentialParameters.compute_signal_at' ...
                     ':Compute the values of the synthetic signal at physical time points.'] ...
                  , {'t:double: array of physical time points'                     , Required ...
                    ,'time_scale:char:''n[ormalized]'' (default) or ''p[hysical]''', 'nb'     ...
                    } ...
                  , varargin{:} ...
                  );
            t = fa.p.t.value;
            % if t is physical time than convert to normalized time: 
            if is_physical(fa.p.time_scale.value)
                t = t*self.sampling_frequency;
            end
            
            bn = transpose(self.b);
            cn = transpose(self.c);
            if size(t,1)~=1
                t = transpose(t);
            end
            signal_values = sum(cn.*bn.^t);
        end
        %------------------------------------------------------------------------------------------                    
        function constructed = construct( self, varargin )
            % [class method] Construct a :class:`Signal` from :class:`MultiExponentialParameters`
            % object.
            %
            % :param nsamples: number of samples in the constructed signal
            %   (a positive integer). 
            % :param key,value=double [t0]: the time point of the first
            %   sample in **physical** time [s]. Default = 0 s. 
            % :param key,value=char [label]: the label of the constructed
            %   :class:`Signal`. Default = 'Synthetic-Signal-NN' 
            %
            % :return: the constructed :class:`Signal` object.
            %
            % `MultiExponentialParameters.construct() examples <examples_MultiExponentialParameters_ctor.html>`_
            %
            fa = Parameters ...
                  ( ['MultiExponentialParameters.construct' ...
                     ':Construct a synthetic signal from the multi-exponential.'] ...
                  , {'nsamples:double: number of samples to generate or an array ', Required ...
                    ,'t0:double: initial time point of the generated signal'      , 0 ...
                    ,'label:char: label of the generated signal'                  , '' ...
                    } ...
                  , varargin{:} ...
                  );
              
             
            nsamples = fa.p.nsamples.value;
            t0       = fa.p.t0.value;
            
            if isempty(fa.p.label.value)
                fa.p.label.set_validated_value(['Synthetic-' Signal.Get_label()]);
            end
            
            if ~nsamples>0
                Sparsimatics.Warning( fa, 'Number of samples is 0.' );
                samples = [];
            else
                nb = length( self.b );
                if ~( nb > 0 )
                    Sparsimatics.Error( fa, 'no base terms found.');
                end
                terms = zeros(nsamples,nb);
                terms(1,:) = self.c .* self.b.^(t0*self.sampling_frequency); % t0 is physical time!
                for it=2:nsamples
                    terms(it,:) = terms(it-1,:) .* self.b ;
                end
                samples = sum(terms,2);
            end
            constructed = Signal( self.sampling_frequency, samples, '--t0',t0, '--label',fa.p.label.value );
        end
      %------------------------------------------------------------------------------------------
% superfluous
%         function [F,A] = frequencies(self)
%             %  [Object method] return the frequencies F [Hz] and optionally
%             %  also the amplitudes [unit unspecified, same as signal]
%             %
%             F = imag(log(self.b)) * self.sampling_frequency / (2*pi);
%             if nargout>1
%                 A = abs(self.c);
%             end
%         end
      %------------------------------------------------------------------------------------------                    
        function ill_disp = ill_disposedness(self,varargin)
            % [Object method] Compute the ill-disposedness of each b.
            %
            % :param key,value=char [algo]: algorithm with which to compute
            %   the ill-disposedness: 'standard' (default), 'undersampled',
            %   'standard2', 'undersampled2', 'poly'.
            % :param key,value=double [parameter]: optional parameter of
            %   the algorithm. 
            % :return: array of ill-disposednesses for each term in the
            %   multi-exponential. 
            %
            % The term Ill-Disposedness appears in the article:
            % B. Beckermann - G.H. Golub - G. Labahn, ON THE NUMERICAL
            % CONDITION OF A GENERALIZED HANKEL EIGENVALUE PROBLEM. 
            %
            fa = Parameters ...
                   ( ['MultiExponentialParameters.ill_disposedness' ...
                      ':Compute ill-disposednes for each base parameter with one of these algorithms: ' ...
                      '''standard'', ''undersampled'', ''standard2'', ''undersampled2'', ''GEP'', ''poly''.'] ...
                   , {'algo:char: algorithm to compute ill-disposedness', 'poly' ...
                     ,'parameter:double: algorithm parameter'           , 1 ...
                     } ...
                   , varargin{:}                    ...
                   );
            algo = fa.p.algo.value;
            parameter = fa.p.parameter.value;
            
            nb = self.nterms;
%             ill_disp = zeros(1,nb);
            signal = self.construct(2*nb);
            H0 = signal.hankel(nb,1);
            H1 = signal.hankel(nb,2);
            HN = norm(H1,2)+norm(H0,2);
            
            if strcmp(algo,'standard')
                % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                V = fliplr(vander(self.b));
                C = diag(self.c);
                W = sqrt(C)*V;
                We = W\eye(nb);
                We2 = dot(We,We);
                ill_disp = (abs(self.b)+1) .* We2;

            elseif strcmp(algo,'undersampled')
                % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                u = parameter;
                W = ones(nb);
                for i = 1:nb
                    W(:,i) = sqrt(self.c)' .* self.b(:).^(u*(i-1)); % da ricontrollare!!!!
                end
                We = W\eye(nb);
                We2 = dot(We,We);
                ill_disp = ( abs(self.b.^u) + 1 ) .* We2;
                
            elseif strcmp(algo, 'poly')
                % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                u = parameter;
                B = self.b.^u;
                ill_disp = ( abs(B) + 1) ./ abs(self.c);
                N = zeros(1,nb);
                for i = 1:nb
                    B_tmp = self.b;
                    B_tmp(i) = [];
                    p = poly(B_tmp);
                    n = polyval(p,self.b(i));
                    p = p./n;
                    N(i) = dot(p,p);
                end
                ill_disp = ill_disp .* N * HN;

            % QUI SOTTO: la differenza sarebbe implementare (vedere se ?? veramente corretto
            % prima) il sistema di Vandermonde per trovare i polinomi e poi, SOLO POI,
            % dividere per i coefficienti delle ampiezze. Questo dovrebbe evitare
            % errori di approssimazione numerica nel calcolare la norma dei polinomi di
            % base di Lagrange

            elseif strcmp(algo, 'standard2')
                % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                V = fliplr(vander(self.b));
                W = V;
                e = eye(nb);
                We = W\e;
                We2 = dot(We,We);
                ill_disp = ( abs(self.b) + 1 ) .* We2 ./ abs(self.c);

            elseif strcmp(algo, 'undersampled2')
                % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                u = parameter;
                V = ones(nb, nb);
                for i = 1:nb
                    V(:,i) = self.b(:).^(u*(i-1)); % da ricontrollare!!!!
                end
                W = V;
                e = eye(nb);
                We = W\e;
                We2 = dot(We,We);
                ill_disp = (abs( self.b .^ u ) + 1) .* We2 ./ abs(self.c);
                % fprintf('cond %d',cond(W));

% the next algorithm is for a signal. it uses GEP to compute the base terms and than the 'standard'
% method to compute the ill-disposedness. this is pointless here.

%             % Potrebbe essere calcolata tramite GEP invece che generata come
%             % Vandermonde matrix, quale sarebbe la differenza??
%             elseif strcmp(algo, 'GEP')
%                 % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
%                 % Codice copiato dal metodo genEigSquare
%                 len = length(self.samples);
% 
%                 if mod(len,2) ~= 0
%                     disp('WARNING, Generalized Eigenvalue Square needs even number of samples...')
%                     disp('discarding last sample')
%                     f = self.samples(1:end-1);
%                 else
%                     f = self.samples(1:end);
%                 end
% 
%                 H0 = hankel(f(1:end/2), f(end/2:end-1));
%                 H1 = hankel(f(2:end/2+1), f(end/2+1:end));
% 
%                 [W,~] = eig(H1,H0);
% 
%                 if sum(isinf(self.b)) ~= 0
%                     disp('WARNING Inf in b!!!!!!!')
%                     self.b(isinf(self.b)) = [];
%                 end
% 
%                 ill_disp = zeros(size(self.b));
% 
%                 for i = 1:nb
%                     e = zeros(nb,1);
%                     e(i) = 1;
%                     ill_disp(i) = ( abs(self.b(i)) + 1) * (norm( W \ e, 2).^2);
%                 end
            else
                Sparsimatics.Error(fa,'Unknown algorithm to compute ill-disposedness: ''%s''.',algo); 
            end
        end
        %------------------------------------------------------------------------------------------
        function reorder(self,I)
            % [object method] Reorder the terms according to the sort order
            % I. So::
            %
            %   self.b = self.b(I)
            %   self.c = self.c(I)
            %
            % :param I: vector of integers. Typically, a permutation of
            %   1:self.nterms. 
            %
            % Note that this function can be used to create a subset of the
            % terms. E.g. if this :class:`MultiExponentialParameters`
            % object has 4 terms and I = [1 3 2], terms 2 and 3 will be
            % interchanged and term 4 will be discarded. 
            
            self.b = self.b(I);
            self.c = self.c(I);         
        end
        %------------------------------------------------------------------------------------------
        function [a_crlb,d_crlb,f_crlb,p_crlb] = crlb( self, varargin )
            % [Object method] compute the Cramer-Rao lower bound (CRLB) of
            % the physical paramters (amplitude, damping, frequency, and
            % phase). The (diagonal of the) Q-matrix is stored in a
            % persistent variable as well as the number of samples. The
            % Q-matrix is only recomputed when *nsamples* changes. 
            %
            % :returns: 
            %   a_crlb, the CRLB of the amplitudes; 
            %   d_crlb, the CRLB of the damping factors;
            %   f_crlb, the CRLB of the frequencies or angular frequencies; 
            %   p_crlb, the CRLB of the phases.
            % :param sigma_noise: variance of the noise in the signal.
            % :param nsamples: number of samples for which the CRLB must be
            %   computed.
            % :param force_q: force computation of the Q-matrix
            %
            fa = Parameters('MultiExponentialParameters.crlb(...)' ...
                   , { 'sigma_noise:@is_strictly_positive:standard deviation of the noise vector' , Required ...
                     , 'nsamples   :@is_strictly_positive:number of samples.'           , Required ...
                     , 'force_q    :logical:force computation of the Q matrix'          , false ...
                     } ...
                   , varargin{:} ...
                 );
            sigmaepsilon2 = fa.p.sigma_noise.value^2;
            
            persistent Qii nsamples a
            
            if isempty(Qii) ...
            || fa.p.force_q.value ...
            || fa.p.nsamples.value~=nsamples
                [a,d,f,p] = MultiExponential.Convert_bc_to_adfp(self.b,self.c);
                nsamples = fa.p.nsamples.value;
                Qii = q_matrix_crlb(a,d,f,p,nsamples);
            end
            
            M = self.nterms;
            
            % SNR_i = eta_i^2 / sigmaepsilon2
            % 1/SNR_i = sigmaepsilon2/eta_i^2
            sigmaepsilon2_over_eta2 = sigmaepsilon2./(a.^2);
            f_crlb = sigmaepsilon2_over_eta2 .* Qii(1,:); % (14)
            d_crlb = sigmaepsilon2_over_eta2 .* Qii(2,:); % (15)
            a_crlb = sigmaepsilon2           .* Qii(3,:); % (16)
            p_crlb = sigmaepsilon2_over_eta2 .* Qii(4,:); % (17)
        end
        %------------------------------------------------------------------------------------------
    end
end
%------------------------------------------------------------------------------------------
% local functions
%------------------------------------------------------------------------------------------
function Qii = q_matrix_crlb( amplitude, damping, frequency, phase, nsamples )
    % [function] compute the Q matrix needed for the Cramer-Rao lower bound
    % of the physical multiexponential parameters for a given number of
    % samples *nsamples*. The implementation follows the paper 
    %
    %   Ying-Xian Yao and Sudhakar M. Pandit (1995) Cramer-Rao Lower Bounds
    %   for a Damped Sinusoidal Process. IEEE Transactions on Signal
    %   Processing.43(4):878-886
    %
    % :param amplitude: amplitudes, real array
    % :param damping:   damping factors [1/s], real array 
    % :param frequency: angular frequency [radians/s], real array
    % :param phase:     phases [radians], real array 
    % :param nsamples:  number of samples
    % :returns: the diagonal of matrix Q, reshaped into a 4 x nterms
    %   matrix. The rows correspond to resp. the angular frequencies, the
    %   damping factors, the amplitudes, and the phases.
    %
    % This function is used by :method:`MultiExponentialParameters.crlb`.
    %

    % translate the input variable names into the symbols used in the
    % paper:
    eta   = amplitude;
    zeta  = damping;
    omega = frequency;
    theta = phase;
    
    M = length(omega);

    A = eta .* exp( 1i .* theta );
    
    % Here we depart a little from the notation in the paper: we assume
    % that the sign of zeta is absorbed in zeta to be consistente with the
    % Sparsimatics conventions.
    z = exp( zeta + 1i.*omega);

    N = nsamples;
    samples = zeros(1,N);
    for s = 1:N
        samples(s) = A(:).' * ( z(:).^(s-1) ) ;
    end

    Zn  = zeros(M, N);
    Zn1 = zeros(M, N);
    Theta = diag( exp( 1i * theta ) );

    for i = 1:N
        Zn( :, i ) = z(:).^ ( i-1 );
        Zn1(:, i ) = (i-1) * z(:) .^( i - 1 );
    end

    Z = [ 1i * Theta * Zn1 ; 
              -Theta * Zn1 ;
               Theta * Zn  ;
          1i * Theta * Zn  ];

    Q = inv( 2 * real( Z * Z' ) );
    Qii = transpose(reshape(diag(Q),M,4));
end
%------------------------------------------------------------------------------------------
