classdef MultiExponentialSolver < handle
    % A wrapper class around the function :meth:`MultiExponential.Solve` that stores the
    % :class:`BSolver` and :class:`CSolver` objects for reusing the same solver parameters several
    % times. 
    %
    % For examples on how to use this class see `tutorial_MultiExponential`_.
    %
    
    %------------------------------------------------------------------------------------------            
    properties 
        parms
    end
    %------------------------------------------------------------------------------------------            
    methods
        %------------------------------------------------------------------------------------------            
        function self = MultiExponentialSolver( varargin )
            % [Constructor] Store a BSolver and a CSolver. 
            % 
            % :param key,value,BSolver [bsolver]: a derived :class:`BSolver` that will produce the normalized base parameters. Optional, default is a :class:`BSolverGEP` object.
            % :param key,value,CSolver [csolver]: a derived :class:`CSolver` that will produce the coefficients associated with the base parameters. Optional, default is a :class:`CSolverVandermondeSquare` object.
            % :param key,value,double [nsamples]: (positive integer or in ]0,1[) set upper limit to the number of samples of the input signal that the solver can use. 
            %
            % Because BSolvers and CSolvers have handle semantics, this method stores a copy of the
            % object, rather than of the handle, to avoid inadvertently modifying its solvers from
            % outside. 
            %
            % There is also a delayed constructor, :meth:`MultiExponentialSolver.init`, which can be
            % use to replace the BSolver and/or the CSolver after construction. 
            %
            self.init( varargin{:} );
        end
        %------------------------------------------------------------------------------------------            
        function self = init( self, varargin )
            % [object method] Set a new BSolver and/or CSolver. 
            %
            % This is a delayed constructor, with the same signature as the constructor
            % :meth:`MultiExponentialSolver.MultiExponentialSolver`. 
            %
            self.parms = Parameters( 'MultiExponentialSolver.init:delayed constructor' ...
                           , {'bsolver:BSolver: BSolver to be used', MultiExponentialSolver.Default_bsolver() ...
                             ,'csolver:CSolver: CSolver to be used', MultiExponentialSolver.Default_csolver() ...
                             ,'nsamples:@is_minus1_or_fraction_or_positive_integer: number of samples to use.', -1 ...
                             } ...
                           , varargin{:} ...
                           );
            
            self.propagate_nsamples_()

            Sparsimatics.Warnv( 1, '! Info', self.parms.get_signature ...
              ,[    'Using a %s object (copy) to solve for the base terms.' ...
                '\n! Using a %s object (copy) to solve for the coefficients.'] ...
              , class(self.parms.p.bsolver.value) ...
              , class(self.parms.p.csolver.value) ...
              );
        end
        %------------------------------------------------------------------------------------------            
        function rv = bsolver(self)
            rv = self.parms.p.bsolver.value;
        end
        %------------------------------------------------------------------------------------------            
        function rv = csolver(self)
            rv = self.parms.p.csolver.value;
        end
        %------------------------------------------------------------------------------------------            
        function [parameters,samples_used] = solve( self, varargin )
            % [object method] This method estimates the (normalized) base
            % terms and coefficients from a :class:`Signal`. 
            %
            % :param signal: a :class:`Signal` object.
            % :return: a :class:`MultiExponentialParameters` object.
            %
            self.parms.add_parameters( {'signal:Signal', Required }, varargin{:} );
            self.propagate_nsamples_()
            if nargout>=2
                % samples_used was requested:
                if isa(self.bsolver,'BSolverComposite')
                    [b,samples_used] = self.bsolver.solve(self.parms.p.signal.value);
                else
                    b = self.bsolver.solve(self.parms.p.signal.value);
                    samples_used = 1:self.bsolver.gep.nsamples;
                end
            else
                b = self.bsolver.solve(self.parms.p.signal.value);
            end
            if isa(self.bsolver,'BSolverSubsampling')
                % don't use all samples, only the the first half of the
                % subsamples. 
                u = self.bsolver.parms.p.rate.value;
                c = self.csolver.solve(self.parms.p.signal.value,b,'--delta',u,'--nsamples',floor(self.parms.p.signal.value.length/2));
            else
                c = self.csolver.solve(self.parms.p.signal.value,b);
            end
% todo: clean up?
%             if isa( self.bsolver, 'BSolverUndersampling' )
%                 u_samples = fa.p.signal.value.undersample( self.bsolver.undersampling_factor, self.bsolver.nsamples_hint ).';
%                 vdm = vandermonde_jk( b, length(u_samples), 0, self.bsolver.undersampling_factor );
%                 c = ( vdm \ u_samples ).';
%             else
%                 copy_from_bsolver = self.csolver.nsamples_hint==CSolver.NsamplesPolicy_CopyFromBSolver;
%                 if copy_from_bsolver
%                     self.csolver.nsamples_hint(self.bsolver.nsamples_hint);
%                 end
%                 c = self.csolver.solve( fa.p.signal.value, b );
%                 if copy_from_bsolver % reset to get the same behavior during the next run.
%                     self.csolver.nsamples_hint(CSolver.NsamplesPolicy_CopyFromBSolver);
%                 end
%             end
            parameters = MultiExponentialParameters ...
                            ( self.parms.p.signal.value.sampling_frequency ...
                            , {b,c}, 'normalized' ...
                            );
%             if exist('samples_used','var')
%                 parameters.samples_used = samples_used;
%             end
        end
        %------------------------------------------------------------------------------------------            
    end    
    %------------------------------------------------------------------------------------------            
    methods(Access=protected)
        %------------------------------------------------------------------------------------------            
        function propagate_nsamples_(self)
            % [protected object method] pass 'nsamples parameter to bsolver
            % and csolver'.
            if ~isa(self.bsolver,'BSolverComposite')
                self.bsolver.parms.p.nsamples.set_validated_value(self.parms.p.nsamples.value)
                self.csolver.parms.p.nsamples.set_validated_value(self.parms.p.nsamples.value)
            end
        end
        %------------------------------------------------------------------------------------------            
    end    
    %------------------------------------------------------------------------------------------            
    methods(Static)
        %------------------------------------------------------------------------------------------            
        function rv = Default_bsolver( varargin )
            % [class method] get/set the default BSolver in Sparsimatics. 
            %
            % :param BSolver [new_DefaultBSolver]: The :class:`BSolver` object that will be used as the default when a new :class:`MultiExponentialSolver` is created. 
            % 
            % Usage::
            %
            %   my_default_bsolver = BSolverEsprit('--nterms',128,'--nrows',BSolver.NrowsPolicy_Square); 
            %   MultiExponentialSolver.DefaultBSolver(my_default_bsolver ); % install the new default BSolver
            %
            % .. note:: This class method stores a handle of the BSolver, 
            %           but returns a **copy** to the stored object.
            %           Consequently, you can modify the default BSolver by
            %           modifying the original object, but it will never be
            %           modified by the internals of
            %           MultiExponentialSolver.
            %
            print_bsolver = false;
            persistent default_bsolver_;
            if nargin==0 && isempty(default_bsolver_)
                default_bsolver_ = BSolverEsprit();
                print_bsolver = true;
            end
            if nargin>0 % reset default_bsolver_
                bsolver = varargin{1};
                assert( isa(bsolver,'BSolver')...
                      ,'MultiExponentialSolver:Default_bsolver(): argument must be a BSolver.')
                default_bsolver_ = bsolver;
                print_bsolver = true;
            end
            if print_bsolver && Sparsimatics.Verbosity()>=3
                cprintf( 'SystemCommands' ...
                       , ['! Info\n' ...
                         ,'! MultiExponentialSolver:Default_bsolver:\n' ...
                         ,obj2str(default_bsolver_) ...
                         ]...
                       );
            end
            rv = copy(default_bsolver_);
        end
        %------------------------------------------------------------------------------------------            
        function rv = Default_csolver( varargin )
            % [class method] get/set the default CSolver in Sparsimatics. 
            %
            % :param CSolver [new_default_csolver]: The :class:`CSolver` object that will be used as the default when a new :class:`MultiExponentialSolver` is created. 
            % 
            % Usage::
            %
            %   my_default_csolver = CSolverVandermondeLS('--nterms',128,'--nrows',CSolver.NrowsPolicy_Square); 
            %   MultiExponentialSolver.DefaultCSolver(my_default_bsolver ); % install the new default CSolver
            %
            % .. note:: This class method stores a handle of the CSolver, 
            %           but returns a **copy** to the stored object.
            %           Consequently, you can modify the default CSolver by
            %           modifying the original object, but it will never be
            %           modified by the internals of
            %           MultiExponentialSolver.
            %
            print_csolver = false;
            persistent default_csolver_;
            if nargin==0 && isempty(default_csolver_)
                default_csolver_ = CSolverVandermondeLS();
                print_csolver = true;
            end
            if nargin>0 % reset verbosity_
                csolver = varargin{1};
                assert( isa(csolver,'CSolver')...
                      ,'MultiExponentialSolver:DefaultCSolver(): argument must be a CSolver.')
                default_csolver_ = csolver;
                print_csolver = true;
            end
            if print_csolver && Sparsimatics.Verbosity()>=3
                cprintf( 'SystemCommands' ...
                       , ['! Info\n' ...
                         ,'! MultiExponentialSolver:Default_csolver:\n' ...
                         ,obj2str(default_csolver_) ...
                         ]...
                        );
            end
            rv = copy(default_csolver_);
        end
        %------------------------------------------------------------------------------------------            
    end
    %------------------------------------------------------------------------------------------            
end

