classdef Signal < handle 
    % A class for storing and manipulating a discrete signal sampled at equidistant points.
    %
    % Class *Signal* stores the sampling frequency [Hz] and the 
    % signal values as a row vector [arbitrary units], which may be 
    % real or complex. It can also accommodate a noise vector. 
    %
    properties
        label='' % [object property] A string for identification purposes, e.g. in plot legends. 
    end
    
    properties (SetAccess=private)
        sampling_frequency  = -1 % [Read-only property] The sampling frequency of the sample.
        samples             = [] % [Read-only property] The signal's sample values. 
        noise               = [] % [Read-only property] The signal's noise values.
        t0                  =  0 % [Read-only property] The (physical time point of the first sample [s].
    end
        
    methods
        %------------------------------------------------------------------------------------------            
        function self = Signal( varargin )
            % Signal constructor. 
            %
            % :param sampling_frequency: the sampling frequency [Hz]. Must be a strictly positive real number.
            % :param samples: row vector of samples, complex or real.
            % :param key,value=double [noise]: row vector of noise, complex or real.
            % :param key,value=double [t0]: the time offset of the first sample [s]. Default = 0.
            % :param key,value=logical [add-noise]: if true (=default), the noise is added to the samples, otherwise the samples are not modified and it is assumed that the noise is already included in the samples.
            % :param key,value=char [label]: the *Signal*'s *label*. Default is automatic label.
            %
            % :return: self.
            %
            % There is also a delayed constructor: *init()* which takes the same parameters and
            % re-initializes the *Signal* object.
            %
            % `Signal() examples <examples_Signal_ctor.html>`_ 
            %
            self.init( varargin{:} );
        end
        %------------------------------------------------------------------------------------------            
        function self = init( self, varargin )
            % [object method] Delayed constructor. Its signature is the same as that of the :meth:`Signal.Signal` constructor.
            %
            % `Signal.init() examples <examples_Signal_ctor.html>`_ 
            %
            fa = Parameters( 'Signal.init:delayed constructor' ...
                   , {'sampling_frequency:@is_strictly_positive: [Hz]'           , Required  ...
                     ,'samples:double: array of sample values'                   , Required  ...
                     ,'noise  :double: noise vector'                             , [] ...
                     ,'t0:double: first time point of the signal'                , 0  ...
                     ,'add_noise:logical: add the noise to the signal, or just store it', true ...
                     ,'label:char: the signal''s label'                          , '' ...
                     } ...
                   , varargin{:} ...
                   );

            self.sampling_frequency = fa.p.sampling_frequency.value;
            
            if isempty( fa.p.label.value )
                self.label = Signal.Get_label();
            else
                self.label = fa.p.label.value;
            end

            nsamples = 0;
            if ~isempty(fa.p.samples.value)
                nsamples = length(fa.p.samples.value );
                self.samples = reshape( fa.p.samples.value, [1,nsamples] );
            end
            if ~isempty(fa.p.noise.value)
                if ~( nsamples == numel(fa.p.noise.value) )
                    Sparsimatics.Error( fa, 'The samples vector and the noise vector must have the same length.');
                end
                self.noise = reshape( fa.p.noise.value, [1,nsamples] );
                if fa.p.add_noise.value
                    self.samples = self.samples+self.noise;
                end
            end
        end
        %------------------------------------------------------------------------------------------                    
        function rv = copy( self, varargin )
            % [object method] Create and return a (shallow) copy of a
            % Signal object, but with a unique label.
            %
            % :param key,value=char [label]: the label of the copy. If not
            %   specified, a unique label is generated. 
            % :return: a copy of self. 
            %
            % `Signal.copy() examples <examples_Signal_ctor.html>`_ 
            %
            fa = Parameters( 'Signal.copy:Create a copy of this signal.' ...
                   , {'label:char: the copy''s label', ''} ...
                   , varargin{:} ...
                   );
        
            if isempty(fa.p.label.value)
                label_of_copy = [self.label ' (copy)'];
            else
                label_of_copy = fa.p.label.value;
            end
        
            rv = Signal( self.sampling_frequency, self.samples ...
                       , '--noise'    , self.noise    ...
                       , '--t0'       , self.t0       ...
                       , '--add-noise', false         ... % just copy the noise vector
                       , '--label'    , label_of_copy ...
                       );
        end
        %------------------------------------------------------------------------------------------            
        function rv = length( self )
            % [object method] Return the number of samples in the signal.
            % 
            % `Signal.length() examples <examples_Signal_misc.html>`_ 
            %
            rv = size( self.samples, 2 );
        end
        %------------------------------------------------------------------------------------------            
        function rv = duration(self)
            % [object method] Compute and return the duration of the signal.
            %
            % `Signal.duration() examples <examples_Signal_misc.html>`_ 
            %
            rv = self.length()/self.sampling_frequency;
        end
        %------------------------------------------------------------------------------------------            
        function rv = sampling_interval( self )
            % [object method] Compute the sampling interval (time between two successive samples) of
            % the signal [s]. 
            %
            % `Signal.sampling_interval() examples <examples_Signal_misc.html>`_ 
            %
            rv = 1/self.sampling_frequency;
        end
        %------------------------------------------------------------------------------------------            
        function rv = energy( self )
            % [object method] Compute and return the energy of the signal. For a discrete signal
            % the `energy <https://en.wikipedia.org/wiki/Energy_(signal_processing)>`_ is defined as
            % the sum of the squared magnitudes of the samples. 
            %
            % `Signal.energy() examples <examples_Signal_misc.html>`_ 
            %
            rv = self.samples * self.samples';
        end
        %------------------------------------------------------------------------------------------            
        function rv = times(self, varargin)
            % [object method] Get an equidistant time array corresponding with the same length as
            % the signal's _samples_ array. 
            %
            % :param key,value=char [time-scale]: the time scale of the time points array, 'p[hysical]' (=default) or 'n[ormalized]'.
            % :return: array with the physical or normalized time points corresponding to *self.samples*.
            %
            % `Signal.times() examples <examples_Signal_misc.html>`_ 
            %
            rv = Signal.Times( self.sampling_frequency, self.length(), self.t0, varargin{:} );
        end
        %------------------------------------------------------------------------------------------            
        function self = add_noise( self, varargin )
            % [object method] Add a noise vector to the signal. If the
            % signal contained already some noise, it is accumulated.
            %
            % :param noise: row vector of noise, complex or real.
            % :return: self.
            %
            % `Signal.add_noise() examples <examples_Signal_noise.html>`_ 
            %
            fa = Parameters('Signal.add_noise:add noise to this signal.' ...
                   , {'noise:double: noise vector', Required} ...
                   , varargin{:} ...
                   );
            
            nsamples = self.length();
            if ~( nsamples == numel( fa.p.noise.value ) )
                  Sparsimatics.Error( fa, 'The samples and the added noise must have the same length.')
            end
            noise_ = reshape( fa.p.noise.value,[1,nsamples] );
            
            self.samples = self.samples + noise_;
            
            if isempty( self.noise )
                self.noise = noise_;
            else
                self.noise = self.noise + noise_;
            end
        end
        %------------------------------------------------------------------------------------------            
        function self = remove_noise( self )
            % [object method] If the signal contains noise, it is removed from the signal. On
            % completion its *noise* property returns an empty array.
            %
            % :return: self
            %
            % `Signal.remove_noise() examples <examples_Signal_noise.html>`_ 
            %
            if ~isempty( self.noise )
                self.samples = self.samples - self.noise;
                self.noise = [];
            end
        end
        %------------------------------------------------------------------------------------------            
        function sigma_noise = add_white_gaussian_noise( self, varargin )
            % [object method] Add white Gaussian noise to the signal values
            % for a given signal to noise ratio snr_. the variance of the
            % noise is returned.
            % If there is a pre-existing noise vector, and it must be
            % replaced, call remove_noise first.
            % 
            % :return: the variance of the noise added.
            % :param snr_: signal to noise ratio.
            % :param key,value=char scale: 'lin'|'log'|'db'.
            %
            % `Signal.add_white_gaussian_noise() examples <examples_Signal_noise.html>`_ 
            %
            
            fa = Parameters( 'Signal.add_white_gaussian_noise' ...
                   , { 'snr:double:signal to noise ratio', Required ...
                     , 'scale:char:''lin''|''log''|''db'''              , 'lin' ...
                     } ...
                     , varargin{:} ...
                   );
            if fa.p.snr.value==Inf
                sigma_noise = 0;
                return
            end
            % snr_ to linear scale
            if strcmp(fa.p.scale.value,'lin')
                snr_ = fa.p.snr.value;
            elseif strcmp(fa.p.scale.value,'log')
                snr_ = 10^fa.p.snr.value;
            elseif strcmp(fa.p.scale.value,'db')
                snr_ = 10^(fa.p.snr.value/10);      
            else
                Sparsimatics.Error( fa.get_signature ...
                  ,' The ''--scale'' parameter must be ''lin''|''log''|''db'', got ''%s''.'...
                  , fa.p.scale.value ...
                  );
            end
            nsamples = self.length;
            
            % (Implementation below after Mathuranathan Viswanathan: 
            % http://www.gaussianwaves.com/2015/06/how-to-generate-awgn-
            % noise-inmatlaboctave-without-using-in-built-awgn-function/). 
            
            % Calculate actual symbol energy:
            energy = sum(abs(self.samples).^2)/nsamples;
            % Find the noise spectral density:
            N0 = energy/snr_;
            if N0<eps
                Sparsimatics.Warnv(1,'! Info', fa.get_signature ...
                    , ' Noise spectral density smaller than eps=%8.2g. No noise added.' ...
                    , eps ...
                    );
                sigma_noise = 0;
                return
            end
            if(isreal(self.samples))
                % Standard deviation for the gaussian white noise when samples is real
                sigma_noise = sqrt(N0);
                wg_noise = sigma_noise*randn(1,nsamples);%computed noise
            else
                % Standard deviation for the gaussian white noise when samples is complex
                sigma_noise  = sqrt(N0/2);
                wg_noise = sigma_noise*(randn(1,nsamples)+1i*randn(1,nsamples));%computed noise
            end
            
            self.add_noise(wg_noise);
         
%             without_noise = self.samples; % store original samples
%             % add noise
%             self.samples = awgn( self.samples, signal_to_noise_ratio, 'measured' );
%             % compute the noise added and add it also to thete noise vector
%             if isempty( self.noise )
%                 self.noise = self.samples - without_noise;
%             else
%                 self.noise = self.noise + self.samples - without_noise;
%             end

        end
        %------------------------------------------------------------------------------------------            
        function H = hankel( self, varargin )
            % [object method] Compute and return the Hankel matrix :math:`H_n^{(r)}(k)` as defined
            % in section 3. of Cuyt and Lee, "How sub-sampling can lead to more robustness and
            % higher resolution in parametric spectral analysis". Note, however, that in the paper *r* starts
            % at 0, but since Matlab is used to index arrays starting at 1, we let r start at 1. 
            %
            % :param key,value=double [n]: integer, the default value equals ceil( self.length()/2 ), which is the maximum.
            % :param key,value=double [r]: integer, the default value is 1. 
            % :param key,value=double [k]: integer, the default value is 1.
            %
            % If *n* is larger than the maximum, it is clipped.
            % Set *Sparsimatics.Verbosity>0* to report this.
            %
            fa = Parameters('Signal.hankel:Compute the Hankel matrix of the signal.' ...
                   , {'n:@is_integer&@is_positive', 0 ...
                     ,'r:@is_integer&@is_strictly_positive', 1 ...
                     ,'k:@is_integer&@is_strictly_positive', 1 ...
                     } ...
                   , varargin{:} ...
                   );
            n = fa.p.n.value;
            r = fa.p.r.value;
            k = fa.p.k.value;
            
            if numel(n) > 2
                Sparsimatics.Warn( fa.get_signature ...
                    ,' Function argument ''n'' has too many elements () and is clipped to the first 2.' ...
                    , numel(n) ...
                    );
            end
            
            if numel(n)==1
                n = [n,n];
            end
                                      
            n_max = ceil( self.length()/2 );
            if n(1) == 0 && n(2) == 0
                n = [n_max,n_max];
            elseif n(1) == 0
                n(1) = self.length() - n(2) + 1;
            elseif n(2) == 0
                n(2) = min(self.length() - n(1) + 1,n(1));
            end
            if n(2) > n(1)
                Sparsimatics.Warn( fa.get_signature ...
                    ,' Function argument number of columns ''n=%d'' is clipped to the number of rows %d.' ...
                    , n(2), n(1) ...
                    );
                n(2) = n(1);
            end
            if n(2) > n_max 
                Sparsimatics.Warn( fa.get_signature ...
                    ,' Function argument number of columns ''n=%d'' is clipped to the maximum value (%d) for signal ''%s''.' ...
                    , n(2), n_max, self.label ...
                    );
                n(2) = n_max;
            end
            if n(1) + n(2) -1 > self.length() 
                Sparsimatics.Warn( fa.get_signature ...
                    ,' Function argument number of rows ''n=%d'' is clipped to the maximum value (%d) for signal ''%s''.' ...
                    , n(1), self.length() - n(2) + 1, self.label ...
                    );
                n(1) = self.length() - n(2) + 1;
            end
            
            r0 = r-1; % in the definition r starts from 0
            H = hankel( self.samples(  1 + r0*k : k : 1 + (r0+n(1)-1)*k )...
                      , self.samples(                 1 + (r0+n(1)-1)*k : k : 1 + (r0+n(1)+n(2)-2)*k ) );
        end
        %------------------------------------------------------------------------------------------            
        function [sub_samples,sub_noise] = subsample(self,varargin)
            % [object method] Return an sub-sampled array from this signal.
            % 
            % :param rate: (integer>0) sub-sampling rate, only every rate-th sample is returned.
            % :param key,value=double [t0_shift]: (integer>0) number of samples to skip befor sub-sampling. Optional, default is 0.
            % :param key,value=double [nsamples]: (integer>0) puts a limit on the number of samples of this *Signal* that can be used. Optional, default is all available samples.
            % :param key,value=double [nsamples_out]: (integer>0) number of samples in the result. Optional, default is as many as possible.
            % :return: two row vectors with sub-sampled samples and corresponding noise (optional): f(t0_shift+1), f(t0_shift+1+1*rate), f(t0_shift+1+2*rate), f(t0_shift+1+3*rate), ...
            %
            fa = Parameters('Signal.subsample:Return a sub-sampled samples and noise array.'  ...
                   , { 'rate        :@is_integer&@is_positive: sub-sampling rate.'    , Required ...
                     , 'shift       :@is_integer&@is_positive: shift of t0.'          ,  0       ...
                     ,['nsamples    :@is_minus1_or_fraction_or_positive_integer'                 ...
                                   ':restrict the length of the input samples array.'], -1       ...
                     , 'nsamples_out:@is_integer:length of output signal'             , -1       ...
                     } ...
                     , varargin{:} ...
                   );
            rate         = fa.p.rate.value;
            t0_shift     = fa.p.shift.value;
            nsamples     = fa.p.nsamples.value;
            nsamples_out = fa.p.nsamples_out.value;
        
            nsamples = apply_fraction_or_min(nsamples,self.length);
            
            if nsamples_out==-1
                nend = nsamples;
            else
                nend = t0_shift + 1 + (nsamples_out-1)*rate;
                if nend>nsamples
                    Sparsimatics.Error( fa.get_signature{1} ...
                      ,[   ' Requested number of samples in the output (nsamples_out=%d) cannot be satisfied:'...
                        '\n! insufficient number of samples in Signal %s of length %d.'] ...
                      , nsamples_out, self.label, self.length ...
                      );
                end
            end
            sub_samples = self.samples(t0_shift+1:rate:nend);
            if nargout==2 
                if isempty(self.noise)
                    sub_noise = [];
                else
                    sub_noise = self.noise(t0_shift+1:rate:nend);
                end
            end
        end
        %------------------------------------------------------------------------------------------            
        function indices = get_indices(self,r,s,varargin)
            % [object method] This function convert a subsampling rate *r*,
            % a shift *s* into an array of indices. The length of the
            % output array can be specified as well as the third argument,
            % as well as number of samples that can be used.
            % 
            % :param r: int>0
            % :param s: int>=0
            % :param [n_out]: int>0
            % :param [n_in]:  int>0
            % :return: array of indices [1+s, 1+s+r, 1+s+2*r, 1+s+3*r, ...] 

            nvarargin = numel(varargin);
            
            switch nvarargin
                case 0
                    indices = 1+s : r : length(self.samples);

                case 1
                    n_out = varargin{1};
                    indices = 1+s : r : 1+s+(n_out-1)*r;
                    
                otherwise
                    assert(false, 'too many arguments')
            end                    
        end
        %------------------------------------------------------------------------------------------            
        function [signal,r] = select(self,varargin)
            % [object method] Create a signal from a subset
            % of this signal's samples. Equidistant sampling is assumed.
            % 
            % :param r: (integer>0) sub-sampling rate( only every r-th 
            %   sample is used), or a range (start:step:stop).
            % :param key,value=double [t0-shift]: (integer>0) number of
            %   samples to skip befor sub-sampling. Optional, default is 0. 
            % :param key,value=double [nsamples]: (integer>0) puts a limit
            %   on the number of samples of this *Signal* that can be used.
            %   Optional, default is all available samples.  
            % :param key,value=double [nsamples_out]: (integer>0) number of
            %   samples in the result. Optional, default is as many as
            %   possible.  
            % :param key,value=char[label]: a label for the new signal.
            % :return: a Signal object containing a sub-sampling of this
            %   signal, and, optionally, also the indices of the used samples.
            %
            fa = Parameters ...
                   (['Signal.select'...
                                   ':Return a signal with a selection of the samples [,noise] arrays.'] ...
                   , {['r           :@is_integer&@is_strictly_positive'...
                                   ':sub-sampling rate or a range.']                  , Required ...
                     ,['shift       :@is_integer&@is_positive'...
                                   ':shift of t0.']                                   ,  0 ...
                     ,['nsamples_out:@is_integer'...
                                   ':length of output signal']                        , -1 ...
                     ,['label       :char'...
                                   ':label for the new signal']                       , '' ...
                     ,['quickanddirty:logical'...
                                   ':if true do not subsample the noise vector, do ' ...
                                   'not give a label']                                , false ...
                     } ...
                   , varargin{:} ...
                   );
            r = fa.p.r.value;
            if length(r)==1
                t0_shift     = fa.p.shift.value;
                nsamples_out = fa.p.nsamples_out.value;
                if nsamples_out==-1
                    r = self.get_indices(r,t0_shift);
                else
                    r = self.get_indices(r,t0_shift,nsamples_out);
                end
            end
            
            %sampling frequency
            u = r(2)-r(1);
            Fs = self.sampling_frequency/u;
            
            %samples
            sub_samples = self.samples(r);
            
            if fa.p.quickanddirty.value
                sub_noise = [];
                lbl = '?';
            else
                %noise
                if isempty(self.noise)
                    sub_noise = [];
                else
                    sub_noise = self.noise(r);
                end
                if ~isempty(fa.p.label.value)
                    lbl = fa.p.label.value;
                else
                    if u==1 
                        fmt = '%s(clipped: [%d, %d, %d, ..., %d])';
                    else
                        fmt = '%s(selected: [%d, %d, %d, ..., %d])';
                    end
                    lbl = sprintf(fmt,self.label,r(1),r(2),r(3),r(end));
                end
            end
            signal = Signal( Fs  ... % the sampling frequency changes!
                           , sub_samples                   ...
                           , '--noise'    , sub_noise      ...
                           , '--t0'       , self.t0 + (r(1)-1)/self.sampling_frequency ...
                           , '--add-noise', false          ... % just copy the noise vector
                           , '--label'    , lbl            ...
                           );
        end
        %------------------------------------------------------------------------------------------            
    end

    methods (Static)
        %------------------------------------------------------------------------------------------            
        function t = Times( varargin )
            % [class method] Construct an equidistant time array with the following parameters:
            % 
            % :param sampling_frequency: the sampling frequency [Hz], which is related to the sampling interval = 1/sampling_frequency.
            % :param n_samples: number of samples in the time array.
            % :param key,value=double [t0]: begin point of the time array. ALWAYS physical time [s]!
            % :param key,value=char [time-scale]: 'p[hysical]' (=default) or 'n[ormalized]'.
            % :return: array with the physical or normalized time points.
            % 
            % `Signal.Times() examples <examples_Signal_misc.html>`_ 
            %
            fa = Parameters('Signal.Times::Return an array of time points.' ...
                   , {'sampling_frequency:@is_strictly_positive: inverse of the sampling interval', Required ...
                     ,'nsamples:@is_integer&@is_strictly_positive: number of time points'   , Required ... 
                     ,'t0:double: physical time [s] of initial time point' ,  0 ...
                     ,'time_scale:char: ''p[hysical]'' or ''n[ormalized]''', 'physical' ...
                     } ...
                   , varargin{:} ...
                   );
            sampling_frequency = fa.p.sampling_frequency.value;
            nsamples           = fa.p.nsamples.value;
            t0                 = fa.p.t0.value;
            time_scale         = fa.p.time_scale.value;
            
            if is_physical(time_scale)
                t = linspace( t0, t0+(nsamples-1)/sampling_frequency, nsamples);
            else
                tau0 = t0*sampling_frequency;
                t = linspace( tau0, tau0 + nsamples-1, nsamples );
            end
        end
        %------------------------------------------------------------------------------------------            
        function rv = Uuid(varargin)
            % [class method] provides a number for automatic labeling of Signal objects. For
            % internal use mainly.  
            %
            % :param [new_counter]: reset the counter. After resetting, the uniqeness of automatic Signal labels is no longer guaranteed.
            %
            persistent uuid_;
            if isempty(uuid_)
                uuid_=0;
            end
            if nargin==1
                uuid_ = varargin{1} - 2;
                % on the next call Uuid() will return varargin{:}
            end
                
            uuid_ = uuid_ + 1;
            rv = uuid_;            
        end
        %------------------------------------------------------------------------------------------            
        function rv = Get_label()
            % [class method] Obtain a default label for a new Signal object. 
            % For internal use mainly. 
            %
            % :return: a string
            % 
            rv = sprintf( 'Signal-%d', Signal.Uuid() );
        end
        %------------------------------------------------------------------------------------------            
    end
end

%{
Handle classes vs value classes
===============================
see https://www.mathworks.com/matlabcentral/newsreader/view_thread/256823
Value class methods canNOT change the internal state of value objects
Handle class methods can.
Conclusion: most classes in a OO design will be handle classes

in addition: make properties which are controlled via methods (SetAccess=private), 
they can be read but only modified through controlled methods, e.g. a delayed ctor.
%}
