classdef Sparsimatics
    % Class Sparsimatics stores global sparsimatics variables such as the verbosity. In addition it
    % has some Static methods to generate uniform warnings and error messages
    % 
    % Technically, this is a Matlab Class, but since it only contains Static methods, it behaves as
    % a namespace.
    %
    % To implement global class properties, we use the "static method with a persistent variable"
    % idiom.
    %
        
    methods(Static)
        %------------------------------------------------------------------------------------------            
        function rv = Verbosity( varargin )
            % [class method] get/set the global verbosity in Sparsimatics. The following values are used:
            %
            %   *  0 : only errors are printed. 
            %   *  1 : Warnings are printed about default values used and choices made by
            %   *  10 : print signature when constructed for objects. This provides a rudimentary execution trace. 
            %   *  11 : print signature when constructed for functions. This provides a rudimentary execution trace. 
            %
            % :param [new_verbosity]: new verbosity value (integer).
            % 
            % Usage::
            %
            %   Sparsimatics.verbosity(1); % set verbosity to 1
            %   v = Sparsimatics.verbosity; % v is now 1.
            %     
            persistent verbosity_;
            if isempty(verbosity_)
                verbosity_ = 0;
            end
            if nargin>0 % reset verbosity_
                v = varargin{1};
                assert( isa(v,'double')&&(v==int32(v))...
                      ,'Sparsimatics:verbosity(): Only integer values are accepted.')
                verbosity_ = v;
                cprintf('SystemCommands','\n\nSparsimatics.verbosity is set to %d.',v)
                switch v
                    case 0
                        cprintf('SystemCommands','\n    Only error messages are printed.')
                    case 1
                        cprintf('SystemCommands','\n    Print info on implicit choices by Sparsimatics.')
                    case 10
                        cprintf('SystemCommands','\n    Trace mode (class members only).')
                    case 11
                        cprintf('SystemCommands','\n    Trace mode (also helper function).')                      
                end
                fprintf('\n\n');
            end
            rv = verbosity_;
        end
        %------------------------------------------------------------------------------------------                          
        function Warn( caller, message, varargin )
            % [class method] Produce a warning if ``Sparsimatics.Verbosity>0``.
            %
            % :param caller: string referring to the method that wanted to print a warning.
            % :param msg: string with the message of the warning.
            %
            if Sparsimatics.Verbosity>0
                Sparsimatics.WARN( 'Warning', caller, message, varargin{:} );
            end
        end
        %------------------------------------------------------------------------------------------            
        function Warnv( v, msg_type, caller, message, varargin )
            % [class method] Produce a warning if ``Sparsimatics.Verbosity>=v``.
            %
            % :param v: print a warning if Verbosity>=v.
            % :param msg_type: title of the warning
            % :param caller: string referring to the method that wanted to print a warning.
            % :param message: (format) string with the message of the warning.
            % :param varargin: optional arguments to be formatted with _message_.
            %
            
            if Sparsimatics.Verbosity>=v
                Sparsimatics.WARN( msg_type, caller, message, varargin{:} );
            end
        end
        %------------------------------------------------------------------------------------------            
        function WARN( msg_type, caller, message, varargin )
            % [class method] Produce a warning (anyway).
            %
            % :param msg_type: title of the warning
            % :param caller: string referring to the method that wanted to print a warning.
            % :param message: string with the message of the warning.
            % :param varargin: optional arguments to be formatted with _message_.
            %
            if startsWith(msg_type,"! ")
                indent = '! ';
            else
                indent = '';
            end
            if iscell( caller ) 
                s = sprintf('\n%s%s',indent,caller{1});
                for i = 2:length(caller)
                    s = sprintf('%s\n%s%s',s,indent,caller{i});
                end
            else
                s = sprintf('\n%s%s',indent,caller);
            end
            s = sprintf('%s:%s',msg_type,s);
            message = sprintf('%s\n',message);
            message = Sparsimatics.Message( s, message, varargin{:} );
            cprintf( 'SystemCommands', message )
        end
        %------------------------------------------------------------------------------------------            
        function Error( caller, message, varargin )
            % [class method] Produce an error.
            %
            % :param caller: string referring to the method that wanted to print a warning.
            % :param msg: string with the message of the warning.
            %
            error( Sparsimatics.Message( caller, message, varargin{:} ) );
        end
        %------------------------------------------------------------------------------------------            
        function msg = Message( caller, message, varargin )
            % [class method] a warning or error message.
            %
            % :param caller: string referring to the method or function that wanted to print a warning.
            % :param msg: string with the message of the warning.
            % :param varargin: optional arguments as in fprintf, sprintf, ...
            % :return: message.
            %
            if isa(caller,'Parameters')
                caller = caller.get_signature{1};
            end
            assert_valid( caller , 'char', 'caller' , 'Sparsimatics:Message( caller, message, varargin )');
            assert_valid( message, 'char', 'message', 'Sparsimatics:Message( caller, message, varargin )');
            fmt = '\n%s:\n! %s\n';
            if ~isempty(varargin)
                msg = sprintf( message, varargin{:} ); 
                msg = sprintf( fmt, caller, msg );
            else
                msg = sprintf( fmt, caller, message );
            end
        end
        %------------------------------------------------------------------------------------------            
    end
    
end


