classdef Parameter < handle
    % class for manipulating a single function parameter
    %------------------------------------------------------------------------------------------
    properties(SetAccess=private)
        parent    = []  % parent Parameters object
        name
        key
        typecheck = ''
        brief     = ''
        repeat    = ' '
        value     = []
        default   = []
        nvalues   = 0
        typestr   = ''
    end
    %------------------------------------------------------------------------------------------
    methods
        %--------------------------------------------------------------------------------------
        function self = Parameter(parent, varargin)
            % [object ctor]
            % 
            % :param Parameters parent: parent Parameters object
            % :param varargin{1}: parameter name
            % :param varargin{2}: parameter type description
            % :param varargin{3}: parameter brief description
            % :param varargin{4}: a char: '1'=required parameter,
            %   non-repeating, '+'=required parameter, repeating: one or
            %   more '*'=required parameter, repeating: zero or more
            %   '?'=optional parameter. Default is '1' for required parameter 
            %   and '?' for optional parameter. 
            % :param varargin{end}: last parameter is the default value for
            %   the parameter. For required arguments its value must be a
            %     instance of class Required.
            % 
            
            self.parent = parent;
            
            nvarargin = length(varargin);
            
            % isolate the last parameter
            default_value = varargin{nvarargin};
            nvarargin = nvarargin-1;

            if nvarargin>4
                disp('varargin=')
                disp(varargin)
                Sparsimatics.Error('Parameter ctor(varargin)' ...
                  , 'To many input arguments, expecting 4 at most.' ...
                  );
            end
            
            if nvarargin>=4
                self.repeat = strtrim(varargin{4});
                if isa(default_value,'Required')
                    if ~contains('1+*',self.repeat)
                        disp('varargin=')
                        disp(varargin)
                        Sparsimatics.Error('Parameter ctor(varargin)' ...
                          , 'Argument 4 (repeat) must be ''1''|''*''|''+'' for a required parameter.'...
                          );
                    end
                else
                    if ~strcmp(self.repeat,'?')
                        disp('varargin=')
                        disp(varargin)
                        Sparsimatics.Error('Parameter ctor(varargin)' ...
                          , 'Argument 4 (repeat) must be ''?''|''*''|''+'' for an optional parameter.'...
                          );
                    end
                end
            else
                if isa(default_value,'Required')
                    self.repeat = '1';
                else
                    self.repeat = '?';
                end
            end
                        
            if nvarargin>=3
                self.brief = strtrim(varargin{3});
                if ~ischar(self.brief)
                    disp('varargin=')
                    disp(varargin)
                    Sparsimatics.Error('Parameter ctor(varargin)' ...
                      , 'Argument 3 (brief description) must be character array.' ...
                      );
                end
            end
            
            if nvarargin>=2
                self.typestr = strtrim(varargin{2});
                if ~ischar(self.typestr)
                    disp('varargin=')
                    disp(varargin)
                    Sparsimatics.Error('Parameter ctor(varargin)' ...
                      , 'Argument 2 (typecheck) must be a character array.' ...
                      );
                end
                self.typecheck = strsplit( strrep(self.typestr,' ',''), '|' );
                for i=1:numel(self.typecheck)
                    if contains(self.typecheck{i},'&')
                        self.typecheck{i} = strsplit(self.typecheck{i},'&');
                    end
                end
            end
            
            if nvarargin>=1
                self.name = strtrim(varargin{1});
                if ~ischar(self.name) ...
                || isempty(self.name) 
                    disp('varargin=')
                    disp(varargin)
                    Sparsimatics.Error('Parameter ctor(varargin)' ...
                      , 'Argument 1 (name) must be a non-empty character array.' ...
                      );
                end
                if ~Parameters.Lean_and_mean
                    % verify that name can serve as a matlab variable
                    try
                        test = [self.name ' = ''test'';']; 
                        eval(test)
                    catch
                        disp('varargin=')
                        disp(varargin)
                        Sparsimatics.Error('Parameter ctor(varargin)' ...
                          , 'Argument 1 (name=''%s'') must not contain characters that cannot occur in matlab variables.'...
                          , self.name ...
                          );
                    end
                end
            end
            
            self.key = ['--' strrep(self.name,'_','-')];
            
            if self.is_optional
                self.set_default(default_value);
            end
            
            % create a name field in the parents p struct:
            s = ['self.parent.p.' self.name ' = self;'];
            eval(s)
            
            if Parameters.Trace
                if self.is_optional
                    s = 'optional';
                else
                    s = 'required';
                end
                cprintf('SystemCommands'...
                       ,'\n! Created %s Parameter ''%s'':' ...
                       , s, self.name ...
                       );
                if ~isempty(self.brief)
                    cprintf( 'SystemCommands'...
                           ,'\n!   brief    = %s'...
                           , self.brief ...
                           );
                end
                cprintf( 'SystemCommands'...
                       ,['\n!   typecheck = ''%s''' ...
                         '\n!   repeat    = ''%s''']...
                       , self.typestr, self.repeat ...
                       );
                if self.is_optional
                    cprintf( 'SystemCommands'...
                           ,['\n!   default   = ' obj2str(self.default)] ...
                           );
                end
                fprintf('\n')
            end
        end
        %--------------------------------------------------------------------------------------        
        function tf = is_repeating(self)
            % [object method] Is this a repeating parameter?
            tf = self.repeat=='+' ...
              || self.repeat=='*';
        end
        %--------------------------------------------------------------------------------------        
        function tf = is_optional(self)
            % [object method] Is this an optional parameter?
            %
            % :return: true or false
            %
            tf = self.repeat=='?';
        end
        %--------------------------------------------------------------------------------------        
        function tf = is_required(self)
            % [object method] Is this an optional parameter?
            %
            % :return: true or false
            %
            tf = ~self.is_optional;
        end
        %--------------------------------------------------------------------------------------        
        function tf = validate(self,value,raise)
            % [object method] Test if value stands the typecheck
            %
            % :param value: value to test, a cell array if this
            %   Parameter is repeating, a single value otherwise.
            % :param logical raise: if true, typecheck failure causes an
            %   error to be raised, otherwise false is returned
            % :returns: true (success), false (failure)
            %
            if self.is_repeating
                tf = isa(value,'cell');
                if ~tf
                    return
                end
                % all entries must stand the test
                tf = validate_one_value(value(1),self.typecheck);                
                for i=2:length(value)
                     tf = tf && validate_one_value(value(i),self.typecheck);
                     if ~tf
                         break % answer is false
                     end
                end
            else
                tf = validate_value(value,self.typecheck);
            end
            if ~tf && raise
                Sparsimatics.Error( self.parent.get_signature{1} ...
                  ,[    'Type check failed for parameter ''%s''.' ...
                    '\n!   Expected   : %s' ...
                    '\n!   Encountered: ''%s'', %s.'] ...
                  , self.name ...
                  , self.typestr ...
                  , class(value), obj2str(value) ...
                  );
            end
        end
        %--------------------------------------------------------------------------------------        
        function tf = validate1(self,value)
            % [object method] Test if one single value stands the
            % typecheck.
            %
            % :param value: value to test.
            % :returns: true (success), false (failure)
            %
            tf = validate_value(value,self.typecheck);
        end
        %--------------------------------------------------------------------------------------
        function set_validated_value(self,value)
            % [object method] store a value that already has been
            % validated.
            %
            % :param value: the value to store, 
            if isa(value,'cell')
                self.nvalues = length(value);
            else
                self.nvalues = 1;
            end
            self.value = value;
        end
        %--------------------------------------------------------------------------------------
        function tf = set_value(self,value,varargin)
            % [object method]
            %
            % :param value: the value to set
            % :param logical varargin{1}: raise an error instead of
            %   returning false.
            % :returns: true if the value was accepted and stored, and
            %   false otherwise.
            %
            if ~isempty(varargin)
                raise = varargin{1};
            else
                raise = true;
            end
            if ~Parameters.Lean_and_mean
                tf = self.validate(value,raise);
                if ~tf
                    return
                end
            end
            self.set_validated_value(value)
        end
        %--------------------------------------------------------------------------------------
        function set_default(self,value)
            self.validate(value,true);
            self.default = value;
            self.value   = value;
        end
        %--------------------------------------------------------------------------------------        
        function s = description(self)
            switch self.nvalues
                case 0
                    if self.is_repeating
                        arg = ['{' self.repeat '}'];
                    else
                        arg = '_';
                    end
                case 1
                    arg = arg2str(self.value);
                    if ~self.is_required && self.is_default
                        arg = ['*' arg '*'];
                    end
                otherwise
                    arg = sprintf('{%d items}', self.nvalues );
            end
            s = sprintf('''%s:%s'' = %s, ', self.name, self.typestr, arg );
        end
        %--------------------------------------------------------------------------------------        
        function tf = is_default(self)
            % [object method] test of this Parameter's value is equal to
            % its default value.
            try
                tf = all( self.value(:) == self.default(:) );
            catch
                tf = false;
            end
        end
        %--------------------------------------------------------------------------------------    
        function reset(self)
            % [object method] Reset this Parameter's value to its
            % default value. Only optional arguments can be reset.
            if self.is_optional
                self.set_validated_value(self.default);
            else
                Sparsimatics.Error( self.parent.get_signature ...
                    , 'Cannot reset a required parameter: ''%s"".' ...
                    , self.name ...
                    );                                
            end
        end
        %--------------------------------------------------------------------------------------        
    end
    %------------------------------------------------------------------------------------------
end
%------------------------------------------------------------------------------------------
% local functions
%------------------------------------------------------------------------------------------
function str = arg2str(arg)
    % [local function] Convert an actual parameter to a string for readable, yet concise, function 
    % signatures.
    %
    if ischar(arg)
        % string between quotes
        str = ['''' arg '''']; 
    elseif numel(arg)==1
        if isnumeric(arg)
            % string representation of numeric value
            str = num2str(arg);
        else
            % class name
            str = class(arg);
        end
    else
        % string containing size and class name
        sz = size(arg);
        str = num2str(sz(1));
        for j=2:length(sz)
            str = [ str 'x' num2str(sz(j)) ]; %#ok<AGROW>
        end
        str = [ str ' ' class(arg) ];
    end
end
%------------------------------------------------------------------------------------------
function i = find_cell(cell_array,s)
    % [local function] search a cell array _cell_array_ of strings for a
    % string _s_
    % :returns: the index of the first occurence of _s_ in the cell array,
    %   or 0 if not found.
    for i=1:numel(cell_array)
        if strcmp(cell_array(i),s)
            return
        end
    end
    i = 0;
end 
%------------------------------------------------------------------------------------------

