classdef Parameters < handle & matlab.mixin.CustomDisplay & matlab.mixin.Copyable
    % Class for handling function and class parameters.
    %------------------------------------------------------------------------------------------
    properties
        method
        brief = ''
        parameters = []
        p = struct % field names are parameter names
    end
    properties(SetAccess=private)
        nrequired = 0
        noptional = 0
        ntotal = 0
        signature = ''
%       disambiguation_allowed = true
        swallowing_possible = false
    end
    properties(Access=private)
        required_parameters_are_set_ = false
        parameter_is_set_ = [] 
        modify_ = false
    end
    %------------------------------------------------------------------------------------------
    methods
        %--------------------------------------------------------------------------------------
        function self = Parameters(method_description,parameter_descriptions,varargin)
            if ~ischar(method_description)
                Sparsimatics.Error('Parameters(method_description,parameter_descriptions,parameter_values)' ...
                  , 'Argument ''method_description'' must be a char or cell of char, got ''%s''.' ...
                  , class(method_description) ...
                  );
            end
            delim = find(method_description==':');
            if isempty(delim)
                self.method = method_description(1:end);
            else            
                self.method = method_description(1:delim-1);
                self.brief  = method_description(delim+1:end);
            end
                        
            self.add_parameters(parameter_descriptions)
            
            if isempty(varargin) 
                tf = self.nrequired>0 ...
                   || ( self.nrequired==0 && ~Parameters.Enable_type_and_order_disambiguation );
            else
                tf = isa(varargin{1},'H');
            end
            if tf % we must print the doc-string
                self.print_doc_string_and_terminate; 
            end
                        
            self.set_values(varargin{:})
            
            if Parameters.Trace
                cprintf( 'SystemCommands'...
                       , ['\n! Created Parameters object:' ...
                         '\n!   %s\n'] ...
                       , self.get_signature{1} ...
                       );
            end
        end
        %--------------------------------------------------------------------------------------
        function add_parameters(self,parameter_descriptions,varargin)
            % [object method] Extend the object with additional function
            % parameters. Values must be set with the modify method. This is
            % because we dont want to have to re-initialize the already
            % existing required parameters.
            %
            % :param parameter_descriptions: cell array containing a series
            %   parameter descriptions, each followed by a default value
            %   (optional parameters) or a Required object (required
            %   parameters). 
            % 
            
            if ~isa(parameter_descriptions,'cell')
                Sparsimatics.Error('Parameters.add_parameters(parameter_descriptions)' ...
                  , 'Arguments descriptions must be a cell array.' ...
                  );
            end
            n = length(parameter_descriptions);
            if mod(n,2)
                Sparsimatics.Error('Parameters.add_parameters(parameter_descriptions)' ...
                  , [    'The parameter descriptions must come in description-default pairs,' ...
                  ,  '\n! but the cell array length is ODD. Description-default pairs consist of:' ...
                  ,  '\n! ''arg_name:type[:brief_function_description[:+|*]]'', default_value|Required'] ...
                  )
            end

            new_names = {};
            new_required = [];
            required = self.parameters(1:self.nrequired);
            optional = self.parameters(self.nrequired+1:self.ntotal);
            required_is_set = self.parameter_is_set_(1:self.nrequired);
            optional_is_set = false(1,self.noptional);
            %optional_is_set = self.parameter_is_set_(self.nrequired+1:self.ntotal);
            for j=1:2:length(parameter_descriptions)
                if ~ischar(parameter_descriptions{j})
                    Sparsimatics.Error('Parameters.add_parameters(parameter_descriptions)' ...
                      , 'Expecting a char array for parameter descriptions #{%d}, got ''%s''.' ...
                      , j, class(parameter_descriptions{j}) ... 
                      );
                end
                ctor_args = strsplit(parameter_descriptions{j},':','CollapseDelimiters',false);
                ctor_args{1} = strtrim(ctor_args{1});
                idx = self.contains(ctor_args{1});
                if idx==0
                    % the parameter does not yet exist
                    nnn = length(new_names);
                    for k=1:nnn
                        if strcmp(ctor_args{1},new_names{k})
                            Sparsimatics.Error('Parameters.add_parameters(parameter_descriptions)' ...
                              , 'Duplicate parameter description for parameter ''%s''.' ...
                              , ctor_args{1} ... 
                              );
                        end
                    end
                    new_names{nnn+1} = ctor_args{1}; %#ok<AGROW>
                    
                    ctor_args{length(ctor_args)+1} = parameter_descriptions{j+1}; % concatenate default value
                    funarg = Parameter( self, ctor_args{:} );
                    if funarg.is_required
                        self.nrequired = self.nrequired+1;
                        new_required = [new_required funarg]; %#ok<AGROW>
                        required = [required funarg]; %#ok<AGROW>
                        required_is_set(self.nrequired) = false;
                    else
                        self.noptional = self.noptional+1;
                        optional = [optional funarg]; %#ok<AGROW>
                        optional_is_set(self.noptional) = false;
                    end
                    if Parameters.Trace
                        cprintf( 'SystemCommands'...
                               ,'\n! Parameters ''%s'' added a ''%s'' parameter.\n' ...
                               , self.get_signature{1}, ctor_args{1}  ...
                               );
                    end
                else
%                     new_required = [new_required self.parameters(idx)]; %#ok<AGROW>
                    if Parameters.Trace
                        cprintf( 'SystemCommands'...
                               ,'\n! Parameters ''%s'' already has a ''%s'' parameter.\n' ...
                               , self.get_signature{1}, ctor_args{1}  ...
                               );
                    end
                end
            end
            self.parameters = [required optional];
            self.ntotal = self.nrequired+self.noptional;
            self.parameter_is_set_ = [required_is_set,optional_is_set];
            
            %verify that a repeating parameter is not followed by an
            %parameter that accepts the same type
            for i=1:self.nrequired-1
                if self.parameters(i).is_repeating
                    if intersect( self.parameters(i).typecheck, self.parameters(i+1).typecheck )
                        Sparsimatics.Error( 'Parameters ctor' ...
                            ,[  'Repeated parameters cannot be followed by an parameter of the same type.' ...
                              '\nArgument %d ''%s'' of type ''%s'' is repeated ''%s'' and followed by .' ...
                              '\nArgument %d ''%s'' of type ''%s''.'] ...
                            , i  , self.parameters(i  ).name, self.parameters(i  ).typestr, self.parameters(i).repeat ...
                            , i+1, self.parameters(i+1).name, self.parameters(i+1).typestr ...
                            );
                    end
                end
            end
            
            % If the last required is repeating, and there are optional
            % parameters of the same type, order and type disambiguation
            % must be suppressed.
            if self.nrequired && self.parameters(self.nrequired).is_repeating
                typecheck = self.parameters(self.nrequired).typecheck;
                for i = self.nrequired+1:length(self.parameters)
                    common = intersect( typecheck, self.parameters(i).typecheck );
                    if ~isempty(common)
                        self.warn_swallowing
                        break
                    end
                end
            end
            
            % Since parameters have been added, the signature (if it was
            % computed) is no longer correct.
            self.signature = '';
            
            % treat parameter values (if there are any)
            nv = length(varargin);
            if nv
%                 nnewreq = length(new_required);
%                 if nv<nnewreq
                if nv<self.nrequired
                    Sparsimatics.Error( self.get_signature ...
                      ,'There are %d required parameters, but only %d values were provided.' ...
                      , nnewreq, nv ...
                      );
                end
                modify_args = cell( 1, self.nrequired + nv );
                for i=1:self.nrequired
                    modify_args{1+2*(i-1)} = self.parameters(i).key;
                    modify_args{  2*i    } = varargin{i};
                end
                if nv>self.nrequired
                    modify_args(2*self.nrequired+1:end) = varargin(self.nrequired+1:end);
                end
                self.modify(modify_args{:})
            end
%             if length(fieldnames(self.p))~=self.ntotal
%                 Sparsimatics.Error( self.get_signature ...
%                   ,'There are less fields in self.p than entries in ' ...
%                   , nnewreq, nv ...
%                   );
%             end
        end
        %--------------------------------------------------------------------------------------
        function m = minimum_nargs(self)
            % [object method] Compute the mininum of parameters that must be
            % provided to set_values
            m = 0;
            if self.modify_
                return
            end
            for i=1:self.nrequired
                if self.parameters(i).repeat~='*'
                    m = m+1;
                end
            end
        end
        %--------------------------------------------------------------------------------------
        function m = maximum_nargs(self)
            % [object method] Compute the maxinum of parameters that must be
            % provided to set_values
            m = 0;
            for i=1:self.nrequired
                if self.parameters(i).repeat=='1'
                    m = m+1;
                else
                    m = Inf;
                    return
                end
            end
            m = m + 2*self.noptional;
        end
        %--------------------------------------------------------------------------------------
        function modify(self,varargin)
            % [object method] Modify (overwrite) some of the required or 
            % optional parameters of a Parameters object (self).
            %
            % :param varargin: cell array with the parameters
            %   (including the keys) that must be modified
            %
            % Note that also the required parameters may be modified, by
            % using key value pairs. E.g. if there is a required parameter
            % _foo_ of type double it may be modified as::
            %
            %   fa.modify('--foo',5)
            %
            % If _foo_ is repeating, and you want to assign more values you
            % must use a cell array for the value::
            %
            %   fa.modify('--foo',{5,6,7})
            %
            % The order relative to optional parameters is immaterial.
            
            self.parameter_is_set_(self.nrequired+1:end) = false;
            self.modify_ = true;
            %varargin{:}
            self.set_values(varargin{:});
            for i=1:self.nrequired
                if ~self.parameter_is_set_(i)
                    Sparsimatics.Error( 'Parameters.modify(...)' ...
                        ,[   'Required parameter #%d ''%s'' of type ''%s'' was not set.' ... 
                         '\n!   Call object.modify( ''%s'', <value> ) to correct this error.']...
                        , i ...
                        , self.parameters(i).name ...
                        , self.parameters(i).typestr ...
                        , self.parameters(i).key ...
                        );
                end
            end
        end
        %--------------------------------------------------------------------------------------
        function reset(self)
            % [object method] Reset each Parameter's value to its
            % default value. 
            for i=self.nrequired+1:self.ntotal
                self.parameters(i).reset
            end
        end
        %--------------------------------------------------------------------------------------
        function set_values(self, varargin)
            
            j = 1;
            n = length(varargin);
            if n < self.minimum_nargs
                Sparsimatics.Error( 'Parameters.set_values' ...
                    , 'Need at least %d (required) parameters, got only %d.' ...
                    , self.nrequired, n ...
                    );                                
            end                
            if n > self.maximum_nargs
                Sparsimatics.Error( 'Parameters.set_values' ...
                    , 'Can accept at most %d parameters (= %d required + 2*%d optional), got too many: %d.' ...
                    , self.nrequired + 2*self.noptional, self.nrequired, self.noptional, n ...
                    );                                
            end                
            if ~self.required_parameters_are_set_
                % process required parameters first
                for i=1:self.nrequired
                    switch self.parameters(i).repeat
                        case '1'
                            if j>n
                                Sparsimatics.Error( 'Parameters.set_values' ...
                                    , [   'No value provided for required parameter ''%s'' of type ''%s''.' ...
                                    , '\n! Exactly one value must be provided.'] ...
                                    , self.parameters(i).name, self.parameters(i).typestr ...
                                    );                                
                            end
                            self.parameters(i).set_value(varargin{j},true);
                            j = j+1;
                        
                        case '+'
                            if j>n
                                Sparsimatics.Error( 'Parameters.set_values' ...
                                    , [   'No value provided for required parameter ''%s'' of type ''%s''.' ...
                                    , '\n! At least one value must be provided (repeat=''+'').'] ...
                                    , self.parameters(i).name, self.parameters(i).typestr ...
                                    );                                
                            end
                            j0 = j;
                            while self.parameters(i).validate1( varargin{j} )
                                j = j+1;
                                if j>n
                                    break
                                end
                            end
                            if j0==j
                                Sparsimatics.Error( 'Parameters.set_values' ...
                                  , 'Required parameter ''%s'': ''+'' repetition requires at least one parameter of type ''%s''.' ...
                                  , self.parameters(i).name, self.parameters(i).typestr ...
                                  );                                
                            end
                            self.parameters(i).set_validated_value( varargin(j0:j-1) );
                        
                        case '*'
                            j0 = j;
                            while self.parameters(i).validate1( varargin{j} )
                                j = j+1;
                                if j>n
                                    break
                                end
                            end
                            self.parameters(i).set_validated_value( varargin(j0:j-1) );

                        otherwise
                            Sparsimatics.Error( 'Parameters.set_values' ...
                              , 'Unexpected error: required parameter %s has repeat==''%s''.' ...
                              , self.parameters(i).name, self.parameters(i).repeat ...
                              );
                    end
                    self.parameter_is_set_(i) = true;
                end
            end
            self.required_parameters_are_set_ = true;
            
%             %set default values for optional parameters:
%             for i=1+self.nrequired:self.ntotal
%                 self.parameters(i).set_validated_value( self.parameters(i).default );
%             end
            
            %process the optional parameters, we start at varargin{j}
            while j<=n
                if is_key(varargin{j}) 
                    if j==n
                        Sparsimatics.Error( 'Parameters.set_values' ...
                          , 'Key ''%s'' is not followed by a value.' ...
                          , self.parameters(i).key ...
                          );                                
                    end
                    self.process_value_by_key(varargin{j:j+1})
                    j = j+2;
                else
                    if ~Parameters.Enable_type_and_order_disambiguation
                        Sparsimatics.Error( 'Parameters.set_values' ...
                          ,[    'Parameters.Enable_type_and_order_disambiguation was set to false.' ...
                            '\n! Optional parameters must be specified as key-value pairs'] ...
                          );                        
                    end
%                     if ~self.disambiguation_allowed
%                         Sparsimatics.Error( 'Parameters.set_values' ...
%                           ,[    'The required parameter ''%s'' is repeating (%s).' ...
%                             '\n! Optional parameters must be presented as key-value pairs, since' ...
%                             '\n! correct type and order disambiguation of optional parameters is' ...
%                             '\n! not guaranteed in this situation.'] ...
%                           , self.parameters(self.nrequired).name ... 
%                           , self.parameters(self.nrequired).repeat ... 
%                           );                        
%                     end
                    ok = self.process_value_by_order(varargin{j});
                    if ~ok
                        ok = self.process_value_by_type(varargin{j});
                    end
                    if ~ok
                        switch j
                            case 1
                                s = '1st';
                            case 2
                                s = '2nd';
                            case 3
                                s = '3rd';
                            otherwise
                                s = sprintf('%dth',j);
                        end
                        Sparsimatics.Error( 'Parameters.set_values' ...
                          ,[    'The %s parameter value was rejected by all parameters.' ...
                            '\n!   value = ' obj2str(varargin{j})] ...
                          , s ... 
                          );                        
                    end
                    j = j+1;
                end
            end
        end
        %--------------------------------------------------------------------------------------
        function process_value_by_key(self,key,value)
            % find Parameter of key
            if self.modify_
                first = 1;
            else 
                first = self.nrequired+1;
            end
            for i=first:self.ntotal
                if strcmp(key,self.parameters(i).key)
%                     if (self.parameters(i).is_optional ... % required p may always be overwritten
%                         && ~Parameters.Enable_duplicate_keys ...
%                         && self.parameter_is_set_(i))
%                         
%                         Sparsimatics.Error( 'Parameters.set_values' ...
%                           ,[    'The optional parameter ''%s'' is specified more than once.' ...
%                             '\n! You must set ''Parameters.Enable_duplicate_keys(true)''' ...
%                             '\n! if this is not to be considered as an error.'] ...
%                           , key ... 
%                           );                        
%                     end
                    self.parameters(i).set_value(value,true);
                    self.parameter_is_set_(i) = true;
                    if Parameters.Trace
                        cprintf( 'SystemCommands'...
                               ,['\n! The value of optional Parameter ''%s'' was set to ' obj2str(value) '.' ...
                                 '\n!   (Accepted as key-value pair).\n' ] ...
                               , self.parameters(i).name  ...
                               );
                    end
                    return
                end
            end
            Sparsimatics.Error( 'Parameters.set_values' ...
              , 'Optional parameter ''%s'' is not found in the parameter descriptions.' ...
              , key ...
              );                        
        end
        %--------------------------------------------------------------------------------------
        function tf = process_value_by_order(self,value)
            % [object method] Return true if the next key accepts a value
            % of type *class(value)*. The *value* is accepted for that key.
            %

            % find the first non-empty key
            for i=self.nrequired+1:self.ntotal
                if ~self.parameter_is_set_(i)
                    % test acceptance
                    tf = self.parameters(i).validate(value,false);
                    if tf
                        self.parameters(i).set_validated_value(value);
                        self.parameter_is_set_(i) = true;
                        if Parameters.Trace
                            cprintf( 'SystemCommands'...
                                   ,['\n! The value of optional Parameter ''%s'' was set to ' obj2str(value) ...
                                     '\n!   (Accepted value by parameter order).\n' ] ...
                                   , self.parameters(i).name  ...
                                   );
                        end
                    end
                    % and return the result
                    return
                end 
            end
            % no empty keys - that is an error too!
            Sparsimatics.Error( 'Parameters.set_values' ...
              , [    'Order disambiguation failed because there are no empty parameters left, '...
                 '\n! while processing %s.'...
                 '\n! (perhaps too many parameters?)'] ...
              , obj2str(value)  ...
              );
        end
        %--------------------------------------------------------------------------------------
        function tf = process_value_by_type(self,value)
            % [object method] Return true if there is only one key
            % accepting values of type *class(val)*. The value *val* is
            % accepted for that key. 
            %

            % loop over the non-empty keys, and test if they accept
            for i=self.nrequired+1:self.ntotal
                if ~self.parameter_is_set_(i)
                    tf = self.parameters(i).validate(value,false);
                    if tf
                        self.parameters(i).set_validated_value( value )
                        self.parameter_is_set_(i) = true;
                        if Parameters.Trace
                            cprintf( 'SystemCommands'...
                                   ,['\n! The value of optional Parameter ''%s'' was set to ' obj2str(value) ...
                                     '\n!   (Accepted value by parameter type: ''%s'').\n' ] ...
                                   , self.parameters(i).name, self.parameters(i).typestr ...
                                   );
                        end
                        return
                    end
                end 
            end
            tf = false;
        end
        %--------------------------------------------------------------------------------------
        function s = get_signature(self)
            % [object method] Compute and return the function's signature.
            % For speed it is computed only once and stored. Note that it 
            % is construct as a cell array of strings. Each cell represents
            % a line.
            if isempty(self.signature)
                if self.ntotal==0
                    s = { [ self.method '()'] };
                else
                    s = cell(2+self.ntotal,1);
                    s{1} = self.method;
                    desc = self.parameters(1).description;
                    s{2}       = ['  ( ' desc(1:end-2) ];
                    for i = 2:self.ntotal
                        desc = self.parameters(i).description;
                        s{1+i} = ['  , ' desc(1:end-2) ];
                    end
                    s{2+self.ntotal} = '  )';
                end
                self.signature = s;
            else
                s = self.signature;
            end
        end
        %--------------------------------------------------------------------------------------
        function s = doc_string(self)
            % [object method] Create and return the function's doc-string.
            %
            if self.nrequired==0
                required_parameters = sprintf('\n!   (none)');
            else
                w_name = 0;
                w_type = 0;
                for i=1:self.nrequired
                    w_name = max(w_name,length(self.parameters(i).name));
                    w_type = max(w_type,length(self.parameters(i).typecheck));
                end
                fmt = ['%s\n!   %-' num2str(w_name) 's:%-' num2str(w_type) 's:%s%s'];
                required_parameters = '';
                for i=1:self.nrequired
                    if self.parameters(i).repeat=='1'
                        r = '';
                    else
                        r = [' (repeat=''' self.parameters(i).repeat ''').'];
                    end
                    required_parameters = sprintf( fmt ...
                                                 , required_parameters ...
                                                 , self.parameters(i).name ...
                                                 , self.parameters(i).typestr ...
                                                 , self.parameters(i).brief ...
                                                 , r ...
                                                 );
                end
            end    
            
            if self.noptional==0
                optional_parameters = sprintf('\n!   (none)');
            else
                w_name = 0;
                w_type = 0;
                for i=self.nrequired+1:self.ntotal
                    w_name = max(w_name,length(self.parameters(i).name));
                    w_type = max(w_type,length(self.parameters(i).typecheck));
                end
                fmt = ['%s\n!   %-' num2str(w_name) 's:%-' num2str(w_type) 's:%s\n!     default = %s'];
                optional_parameters = '';
                for i=1+self.nrequired:self.ntotal
                    sdef = obj2str(self.parameters(i).default);
                    optional_parameters = sprintf( fmt ...
                                                , optional_parameters ...
                                                , self.parameters(i).name ...
                                                , self.parameters(i).typestr ...
                                                , self.parameters(i).brief ...
                                                , sdef ...
                                                );
                end
            end
            
            self.format_function_brief;
            s = sprintf( ['\n'                           ...
                          '\nDoc-string for %s:%s'       ...
                          '\n! Required parameters:%s'   ...
                          '\n! Optional parameters:%s'   ...
                          '\n\n']                        ...
                       , self.method                     ...
                       , self.brief                      ...
                       , required_parameters             ...
                       , optional_parameters             ...
                       );
        end
        %-------------------------------------------------------------------------------------
        function format_function_brief(self)
            % [private object method] Format the function's brief explanation.
            if ~isempty(self.brief)
                brf = [self.brief,' '];
                self.brief = '';
                spaces = find(brf==' ');
                s0 = 1;
                n = length(spaces);
                for i1=1:n
                    s1 = spaces(i1);
                    if s1 > s0+75 ...
                    || i1 == n
                        self.brief = [self.brief,'\n!   ',brf(s0:s1)];
                        s0 = s1+1;
                    end
                end
            end
        end
        %--------------------------------------------------------------------------------------
        function print_doc_string(self)
            %todo set doc-string
            cprintf('Comments', self.doc_string );
        end
        %--------------------------------------------------------------------------------------
        function print_doc_string_and_terminate(self)
            self.print_doc_string; 
            Sparsimatics.Error( self.method  ...
                ,'Interrupting program execution after printing the doc-string.' ...
            ); 
        end
        %----------------------------------------------------------------------------------
        function warn_swallowing(self)
            Sparsimatics.Warnv( 3, '! WARNING', self.get_signature ...
              ,['\n! Because there are optional parameters with the same type as a'    ...
                '\n! repeating required parameter, the repeating parameter may swallow' ...
                '\n! optional parameters.'...
                '\n! YOU MUST USE KEY-VALUE PAIRS TO SET OPTIONAL ARGUMENTS UNAMBIGUOUSLY.']...
              );
            self.swallowing_possible = true;
        end
        %----------------------------------------------------------------------------------
        function idx = contains(self,argument_name)
            % [protected object method] Verify if there is already an
            % parameter with this name.
            for idx=1:self.ntotal
                if strcmp(self.parameters(idx).name,argument_name)
                    return
                end
            end
            idx = 0;
        end
        %----------------------------------------------------------------------------------
    end
    %--------------------------------------------------------------------------------------
    methods(Static)
        %----------------------------------------------------------------------------------
        function tf = Lean_and_mean(varargin)
            % [class method] This class method mimicks a class property (by
            % exposing a persistent variable). Lean_and_mean exposes a
            % logical flag. If true most time consuming sanity checks are
            % just skipped.
            % The initial value is *false*. 
            %
            % :param [new_state]: the new state of the flag, logical
            % :return: the current state of the trace flag, logical.
            %
            persistent lean_and_mean_;
            if isempty(lean_and_mean_)
                lean_and_mean_ = false;
            end
            if nargin>0
                if nargin>1
                    Sparsimatics.Error('Parameters:Trace(tf)'...
                      , 'Expecting 0 or 1 parameters, got %d.' ...
                      , nargin ...
                      );
                end
                lean_and_mean = varargin{1};
                if ~isa(lean_and_mean,'logical')
                    Sparsimatics.Error('Parameters:Trace(tf)' ...
                        , 'Expecting a ''logical'' parameter, got ''%s''.' ...
                        , class(trace) ...
                        );
                end
                lean_and_mean_ = lean_and_mean;
            end
            tf = lean_and_mean_;
        end
        %----------------------------------------------------------------------------------
        function tf = Trace(varargin)
            % [class method] This class method mimicks a class property (by
            % exposing a persistent variable). Trace exposes a logical flag
            % that makes the constructor of class Parameters print
            % an execution trace that helps the user to understand how 
            % optional parameters are interpreted. The initial value is
            % *false*. 
            %
            % :param [new_state]: the new state of the flag, logical
            % :return: the current state of the trace flag, logical.
            %
            persistent trace_;
            if isempty(trace_)
                trace_ = false;
            end
            if nargin>0
                if nargin>1
                    Sparsimatics.Error('Parameters:Trace(tf)'...
                      , 'Expecting 0 or 1 parameters, got %d.' ...
                      , nargin ...
                      );
                end
                trace = varargin{1};
                if ~isa(trace,'logical')
                    Sparsimatics.Error('Parameters:Trace(tf)' ...
                        , 'Expecting a ''logical'' parameter, got ''%s''.' ...
                        , class(trace) ...
                        );
                end
                trace_ = trace;
            end
            tf = trace_;
        end
        %------------------------------------------------------------------------------------------            
        function rv = Enable_type_and_order_disambiguation( varargin )
            % [class method] implement a static property 
            %
            % :param tf: *true*, false
            % 
            % Usage::
            %
            %   Parameters.Enable_type_and_order_disambiguation(false);
            %     
            persistent enable_type_and_order_disambiguation_;
            if isempty(enable_type_and_order_disambiguation_)
                enable_type_and_order_disambiguation_ = true;
            end
            if nargin>0 % reset enable_type_and_order_disambiguation_
                v = varargin{1};
                if ~( isa(v,'logical') || v==0 || v==1 )
                    Sparsimatics.Error( 'Parameters.Enable_type_and_order_disambiguation' ...
                      ,'Only logical values are accepted (true|false|0|1), got ''%s''.' ...
                      , class(v) ...
                      );
                end
                enable_type_and_order_disambiguation_ = v;
                cprintf('SystemCommands','\n\nParameters.Enable_type_and_order_disambiguation is set to %d.\n',v)
            end
            rv = enable_type_and_order_disambiguation_;
        end
        %------------------------------------------------------------------------------------------            
%         function rv = Enable_duplicate_keys( varargin )
%             % [class method] implement a static property 
%             %
%             % :param tf: *true*, false
%             % 
%             % Usage::
%             %
%             %   Parameters.Enable_duplicate_keys(false);
%             %     
%             persistent enable_duplicate_keys_;
%             if isempty(enable_duplicate_keys_)
%                 enable_duplicate_keys_ = false;
%             end
%             if nargin>0 % reset enable_duplicate_keys_
%                 v = varargin{1};
%                 if ~( isa(v,'logical') || v==0 || v==1 )
%                     Sparsimatics.Error( 'Parameters.Enable_duplicate_keys' ...
%                       ,'Only logical values are accepted (true|false|0|1), got ''%s''.' ...
%                       , class(v) ...
%                       );
%                 end
%                 enable_duplicate_keys_ = v;
%                 cprintf('SystemCommands','\n\nParameters.Enable_duplicate_keys is set to %d.\n',v)
%             end
%             rv = enable_duplicate_keys_;
%         end
        %--------------------------------------------------------------------------------------
    end
    %------------------------------------------------------------------------------------------
    methods (Access=protected)
        %--------------------------------------------------------------------------------------
        function propgrp = getPropertyGroups(self)
            if ~isscalar(self)
                propgrp = getPropertyGroups@matlab.mixin.CustomDisplay(self);
            else
                s = {};
                for i=1:self.ntotal
                    s{1+2*(i-1)} = self.parameters(i).name;  %#ok<AGROW>
                    s{  2*i    } = self.parameters(i).value; %#ok<AGROW>
                end
                propList = struct(s{:});
                propgrp = matlab.mixin.util.PropertyGroup(propList);
            end
        end
        %------------------------------------------------------------------------------------------
    end
end
%------------------------------------------------------------------------------------------
% local functions
%------------------------------------------------------------------------------------------
function tf = is_key(s)
    tf = false;
    if ~ischar(s) 
        return
    end
    tf = startsWith(s,'--');
end
%------------------------------------------------------------------------------------------
function common = intersect(types1,types2)
    % :param types1: cell array of typechecking instructions
    % :param types2: cell array of typechecking instructions
    %
    % :returns the intersection of types1 and types2 
    % 
    if isempty(types1)
        common = '';
        return 
    end
    common = '';
    for i1=1:length(types1)
        i2 = 1;
        n2 = length(types2);
        while i2<=n2
            if ischar(types1{i1}) ...
            && ischar(types2{i2})
                if strcmp(types1(i1),types2(i2))
                    common = [common '|' types1{i1}]; %#ok<AGROW>
                    types2(i2) = [];
                    break % out of for i2
                end
            else
                if iscell(types1{i1}) ...
                && iscell(types1{i2}) 
                    same = false;
                    for j1=1:numel(types1{i1})
                        found = false;
                        for j2=1:numel(types1{i2})
                            if strcmp(types1{i1}(j1),types2{i2}(j2))
                                found = true; 
                                break
                            end
                        end
                        if ~found 
                            break    
                        end
                        same = true;
                    end
                    if same
                        type_i1 = types1{i1}(1);
                        for j1 = 2:numel(types1{i1})
                            type_i1 = [type_i1 '&' types1{i1}(j1)]; %#ok<AGROW>
                        end
                        common = [common '|' type_i1]; %#ok<AGROW>
                        break % out of while i2<=n2
                    end
                end
            end
            i2 = i2+1;
        end
    end
    if ~isempty(common)
        common = common(2:end); % remove leading '|'
    end
end
%------------------------------------------------------------------------------------------
