function components_to_remove = filter_components(varargin)
    % [function] This function returns a list of components to _remove_ in
    % order to keep only the *n* components corresponding to the first *n*
    % entries after sorting the quantity *q* in the specified *direction*.
    %
    % :returns: an array of indices to remove.
    % :param q: a double array from which only the *n* smallest/largest
    %   components should be kept.
    % :param n: the number of components to keep.
    % :param keep: 'l[argest]' (=default), or 's[mallest]'.
    %
    % Typical use, to keep the n multi-exponential components with the
    % largest energy:: 
    %
    %   >> remove_components = filter_components(multi_exponential_components.energies,n)
    %   >> multi_exponential_components.remove(remove_components)
    %
    
    fa = Parameters(['filter_components: return the indices of n components '...
                                        'with q smaller|greater than the n '...
                                        'largest|smallest q-values.'] ...
           , { 'q:double:quantity to be filtered'                                   , Required ...
             , 'n:@is_integer&@is_strictly_positive: number of components to keep'  , Required ...
             ,['keep:char:''l[argest]''=keep n largest components (default), '...
                         '''s[mallest]''=keep n smallest components.']             , 'l'...
             } ...
           , varargin{:} ...
           );
    
    if startsWith('largest',fa.p.keep.value)
        largest = true;
    elseif startsWith('smallest',fa.p.keep.value)
        largest = false;
    else
        Sparsimatics.Error( fa.get_signature ...
          , 'Parameter keep must be either ''l[argest]'', or ''s[mallest]'', got ''%s''.' ...
          , fa.p.keep.value ...
          );
    end
    n = fa.p.n.value;
    if n>=length(fa.p.q.value)
        components_to_remove = [];
        return
    end
    [~,I] = sort(fa.p.q.value);
    if largest
        components_to_remove = I(1:end-n);
    else
        components_to_remove = I(n+1:end);
    end
end