function figure_handle = plot3_base_terms( varargin )
    % [function] Plot the base parameters *b* in the complex plane and a
    % third dimension. A unit circle is added in the horizontal plane for
    % reference. 
    %
    % :param MultiExponentialParameters p: one or more
    %   *MultiExponentialParameters* objects. 
    % :param key,value=char [z]: selects the value to be plotted on the
    %   z-axis: 'abs': magnitude of *c* coefficient (default), 'angle':
    %   angle of *c* coefficient, 'energy': energy of the components. 
    % :param key,value=double [duration]: when ``'--z'=='energy'`` the
    %   energy will be computed for a signal with a duration of this many
    %   seconds, default = 1s. 
    % :param key,value=cell [legend]: legend entries, one for each p entry.
    % :param key,value=double [width]: if nonzero, sets the plot axis
    %   limits to ``axis( [-width,width,-width,width] );``, otherwise no
    %   axis limits are applied (showing all points plotted).   
    % :param key,value,double [radii]: array of radii to draw, default =
    %   [1], which just draws the unit circle. 
    % :return: a handle to the created figure.
    % 
    % `Examples <tutorial_Visualization.html>`_
    %
    fa = Parameters ...
           ( ['plot3_base_terms' ...
              ':Plot the base terms in the complex plane, and add a '...
              'z-axis for abs(c), angle(c), or the component energy.'] ...
           , {'b:MultiExponentialParameters: base terms and coefficients:+'                                  , Required ...
             ,'z:char: value to be plotted on the z-axis = ''abs''|''angle''|''energy''|''ill-disposedness''', 'abs' ...
             ,'nlargest:double: only the nlargest most energetic components will be plotted (default=0=all)' , 0     ...
             ,'duration:double: duration of the signal used to compute the energy when ''--z''==''energy'''  , 1     ...   
             ,'legend:cell:legend entries'                                                                   , {} ...
             ,'width:double: axis limits'                                                                    , 1.2   ...
             ,'radii:double: array of radii to draw'                                                         , 1 ...
             } ...
           , varargin{:} ...
           );

    figure_handle = figure('Tag' , 'plot3_base_terms' ...
                          ,'Name', 'plot3_base_terms' ...
                          );
    nps = length(fa.p.b.value);
    symbols = repmat('o',nps);
    if nps>1
        symbols(1) = '+';
    end
    
    z = cell(1,nps);
    if strcmp(fa.p.z.value,'abs')
        for i=1:nps
            z{i} = abs(fa.p.b.value{i}.c);
        end
        zlabel_ = 'abs(c)';
    elseif strcmp(fa.p.z.value,'angle')
        for i=1:nps
            z{i} = angle(fa.p.b.value{i}.c);
        end
        zlabel_ = 'angle(c)';
    elseif strcmp(fa.p.z.value,'energy')
        for i=1:nps
            z{i} = fa.p.b.value{i}.energies(fa.p.duration.value ...
                       * fa.p.b.value{i}.sampling_frequency);
        end
        zlabel_ = 'energy';
    elseif strcmp(fa.p.z.value,'ill-disposedness')
        for i=1:nps
            z{i} = fa.p.b.value{i}.ill_disposedness;
        end
        zlabel_ = 'ill-disposedness';
    else
        Sparsimatics.Error(fa,'Invalid z value, expecting ''abs''|''ang''|''energy'', got ''%''.',fa.p.z.value);
    end    
    colors = zeros(nps,3);
    if fa.p.nlargest.value>0
        I = cell(1,nps);
    end
    for i=1:nps
        xi = real(fa.p.b.value{i}.b);
        yi = imag(fa.p.b.value{i}.b);
        zi = z{i};
        if fa.p.nlargest.value>0
            if strcmp(fa.p.z.value,'energy')
                [~,Ii] = sort(zi,'descend');
            else
                Ei = fa.p.b.value{i}.energies(fa.p.duration.value*fa.p.b.value{i}.sampling_frequency);
                [~,Ii] = sort(Ei,'descend');
            end
            Ii = Ii(1:fa.p.nlargest.value);
            I{i} = Ii;
            xi = xi(Ii);
            yi = yi(Ii);
            zi = zi(Ii);
        end
        if i==1
            line = stem3( xi,yi,zi,symbols(i),'Color',[1,0,0] );
        else
            line = stem3( xi,yi,zi,symbols(i) );
        end
        colors(i,:) = get(line,'Color');
        hold on
    end
    % horizontal connections to unit circle
    for i=1:nps
        b = fa.p.b.value{i}.b;
        if fa.p.nlargest.value>0
            Ii = I{i};
            b = b(Ii);
        end
        x = real(b);
        y = imag(b);
        r =  abs(b);
        r(find(r<0.5)) = 1;  %#ok<FNDSB>
        xx1 = [x;x./r];
        yy1 = [y;y./r];
        zz1 = zeros(size(xx1));
        plot3(xx1,yy1,zz1,'-','Color',colors(i,:))
        hold on
    end
    
    if ~isempty(fa.p.legend.value)
        legend(fa.p.legend.value)
    elseif nps>1 
        legend_ = cell(1,nps);
        for i=1:nps
            legend_{i} = num2str(i);
        end
        legend(legend_{:});
    end    
    plot_unit_circle('--width', fa.p.width.value ...
                    ,'--dim'  , 3                   ...
                    ,'--radii', fa.p.radii.value ...
                    );
    zlabel(zlabel_);
    
    % improve daspect
    xl = xlim;
    dx = xl(2)-xl(1);
    zl = zlim;
    dz = zl(2)-zl(1);
    daspect([1,1,dz/dx]);
end
