function figure_handle = plot_base_terms( varargin )
    % [function] Plot the base parameters *b* in the complex plane. A unit
    % circle is added for reference. 
    %
    % :param b: one or more arrays of base terms or
    %   *MultiExponentialParameters* objects. 
    % :param key,value=cell [legend]: legend entries, one for each p entry.
    % :param key,value=double [width]: if nonzero, sets the plot axis
    %   limits to ``axis( [-width,width,-width,width] );``, otherwise no
    %   axis limits are applied (showing all points plotted).
    % :param key,value,double [radii]: array of radii to draw, default = 1,
    %   which just draws the unit circle. 
    % :return: a handle to the created figure.
    % 
    % .. note::
    %
    %    It is recommended to specify the optional arguments as key-value
    %    pairs. Otherwise, they could be accepted as another b parameter,
    %    because it is a repeated argument.
    % 
    % `examples <tutorial_Visualization.html>`_
    %
    fa = Parameters ...
           ( 'plot_base_terms:Plot the base terms in the complex plane.' ...
           , {'b    :double|MultiExponentialParameters: array of base terms:+', Required ...
             ,'legend:cell:legend entries'                                    , {} ...
             ,'width:double: axis limits'                                     , 1.2 ...
             ,'radii:double: array of radii to draw'                          , 1 ...
             } ...
           , varargin{:} ...
          );
    bs    = fa.p.b.value;
    w     = fa.p.width.value;
    radii = fa.p.radii.value;
    
    figure_handle = figure('Tag' , 'plot_base_terms' ...
                          ,'Name', 'plot_base_terms' ...
                          );
    nbs = length(bs);
%     symbols = repmat('o',nbs);
    symbols = cell(1,nbs);
    for i=1:nbs
        symbols{i} = 'o';
        if i==1 && nbs>1
           symbols{i} = '+r';
        end
        if isa(bs{i},'MultiExponentialParameters')
            bs{i} = bs{i}.b;
        end
        % make sure that they are all complex values
        bs{i} = complex(bs{i});
        plot( bs{i}, symbols{i} );
        hold on
    end
    
    plot_unit_circle('--width',w,'--radii',radii);

    if ~isempty(fa.p.legend.value)
        legend(fa.p.legend.value)
    elseif nbs>1 
        legend_ = cell(1,nbs);
        for i=1:nbs
            legend_{i} = num2str(i);
        end
        legend(legend_{:});
    end
    
end
