function figure_handle = plot_signal_SVD( varargin )
    % [function] This function computes the singular value decomposition of a square Hankel matrix
    % from the *signal* 's samples and plots its singular values. 
    %
    % :param signal: a *Signal* object, or 
    % :param key,value=char [plot-what]: a string containing the substrings 'lin' (=plot singular values on linear scale), 'log' (=plot singular values on logarithmic scale). An empty string or '*' causes both to be plotted. 
    % :param key,value=double [n]: integer as in :math:`H_n^{(r)}(k)`, used as input to *signal.hankel()*.
    % :param key,value=double [r]: integer as in :math:`H_n^{(r)}(k)`, used as input to *signal.hankel()*.
    % :param key,value=double [k]: integer as in :math:`H_n^{(r)}(k)`, used as input to *signal.hankel()*.
    % :return: handle of the created figure.
    % 
    % `examples <tutorial_Visualization.html>`_
    %
    fa = Parameters ...
           (['plot_signal_SVD:Compute and plot the singular values ' ...
             'of a square Hankel matrix of the input signal.']        ...
           , {'signal:Signal: signal to plot'                 , Required ...
             ,'plot_what:char: ''lin''|''log''|''*'''         ,'*' ...
             ,'n:@is_integer&@is_positive'                    , 0  ...
             ,'r:@is_integer&@is_strictly_positive'           , 1  ...
             ,'k:@is_integer&@is_strictly_positive'           , 1  ...
             ,'shift:double:shift for each plot if n is array', 0
             } ...
           , varargin{:} ...
           );
    plot_what = fa.p.plot_what.value;
    n         = fa.p.n.value;
    r         = fa.p.r.value;
    k         = fa.p.k.value;
    shift     = fa.p.shift.value;
    
    if n==0
        n=floor(fa.p.signal.value.length/2);
    end
                                
    if isempty(plot_what)
        plot_what = '*';
    end
    plot_all = contains(plot_what,'*');
    if plot_all
        plot_lin= true;
        plot_log= true;
    else
        plot_lin= contains(plot_what,'lin');
        plot_log= contains(plot_what,'log');
        if ~plot_lin && ~plot_log % if nothing was specified, by default all are plotted
            plot_lin = true;
            plot_log = true;
        end
    end

    figure_handle = figure('Tag','singular_values');
    legend_ = {};

    nj=length(n);
    if nj==1
        plot_symbol1 = 'b-';
        plot_symbol2 = 'ro';
    else
        plot_symbol1 = 'o-';
    end
    
    for jj=1:nj
        legend_{jj} = num2str(n(jj)); %#ok<AGROW>
        
        H = fa.p.signal.value.hankel(n(jj),r,k);
        [~,s,~] = svd(H); 
        s = diag(s);
        ls = length(s);
        x = (1:ls) + (jj*shift);

        n_subplots = double( int16(plot_lin)+int16(plot_log) );
        i_subplot = 0;

        n_ = script(n,'_');
        r_ = script(r,'^');
        title_ = sprintf('Singular values of H%s%s(%d) for "%s"',n_,r_,k,fa.p.signal.value.label); 

        if plot_lin
            i_subplot = i_subplot+1;
            subplot(n_subplots,1,i_subplot)
            plot(x,s,plot_symbol1);
            hold on
            if nj==1
                plot(x,s,plot_symbol2);
            end
            title(title_);
            ylabel('y');
%             if j<length(n)
%                 hold on
%             end
        end
        if plot_log
            i_subplot = i_subplot+1;
            subplot(n_subplots,1,i_subplot)
            semilogy(x,s,plot_symbol1)
            hold on
            if nj==1
                semilogy(x,s,plot_symbol2)
            end
            title(title_);
            ylabel('log(y)');
%             if j<length(n)
%                 hold on
%             end
        end
    
    end
    legend(legend_);
end

function out = script(in,c)
    % create superscript string if c=='^' or subscript if c=='_'
    if isnumeric(in)
        in = num2str(in);
    end
    n = length(in);
    
    out = blanks(2*n);
    for i=1:n
        out(2*i-1) = c;
        out(2*i  ) = in(i);
    end
end

