function figure_handle = plot_terms( varargin )
    % [function] This function creates a plot of the separate terms of a
    % multi-exponential series. It creates 3 subplots, resp. for the real
    % part, the imaginary part and the magnitude.  
    %
    % :param parameters: the *MultiExponentialParameters* object to be
    %   plotted. 
    % :param key,value=char [plot_what]: a string containing the characters
    %   'R' (=plot real part), 'I' (=plot imaginary part), 
    %   'M' (=plot magnitude). An empty string or '*' causes everything to
    %   be plotted. Adding an 's' causes also the sum to be plotted.
    % :param key,value=char [time_scale]: a string identifying the time
    %   scale: 'n[ormalized]' or 'p[hysical]'. The default 'physical'.  
    % :param key,value=char [nsamples]:  number of samples to be plotted,
    %   default amounts to a duration of 1s. 
    % :param key,value=char [t0]: initial time point [s].
    %   Default = 0s. 
    % :return: handle of the created figure.
    %
    % `examples <tutorial_Visualization.html>`_
    %
    fa = Parameters ...
           ( 'plot_terms:Plot the individual components of a multi-exponential in the time domain.' ...
           , {['parameters:MultiExponentialParameters: the MultiExpo' ...
               'nentialParameters object whose terms must be plotted']	, Required   ...
             ,'nsamples:double: number of samples to be plotted'        , 0          ...
             ,'plot_what:char: ''RIMs'' '                               , '*'        ...
             ,'time_scale:char: ''p[hysical]''|''n[ormalized]'''        , 'physical' ...
             ,'t0:double: initial time point [s]'                       , 0          ...
             } ...
           , varargin{:} ...
           );
    %[parameters, fa.p.nsamples.value, fa.p.plot_what.value,fa.p.time_scale.value, fa.p.t0.value] = fa.actual_values{:};
    
    physical = is_physical( fa.p.time_scale.value ); % assert that fa.p.time_scale.value is valid.
    if fa.p.nsamples.value==0
        fa.p.nsamples.set_value( fa.p.parameters.value.sampling_frequency ); %-> 1s duration
    end
    
    terms = fa.p.parameters.value.terms(fa.p.nsamples.value, '--t0', fa.p.t0.value );
    signal= sum(terms,2);
    
    figure_handle = figure( 'Name', 'plot_terms' );

    if isempty(fa.p.plot_what.value)
        fa.p.plot_what.value = '*';
    end
    plot_all = ~isempty(strfind(fa.p.plot_what.value,'*'));
    if plot_all
        plot_sg = true;
        plot_re = true;
        plot_im = true;
        plot_mg = true;
    else
        plot_sg = ~isempty(strfind(fa.p.plot_what.value,'s'));
        plot_re = ~isempty(strfind(fa.p.plot_what.value,'R'));
        plot_im = ~isempty(strfind(fa.p.plot_what.value,'I'));
        plot_mg = ~isempty(strfind(fa.p.plot_what.value,'M'));
    end
    
    nterms_ = fa.p.parameters.value.nterms;
    legend_ = {};
    if plot_sg
        d=1;
        legend_{d} = 'signal';
    else
        d=0;
    end
    for i=1:nterms_
        legend_{d+i} = sprintf('i=%d',i); %#ok<AGROW>
    end

    if physical
        xlabel_ = 'time [s]';
    else        
        xlabel_ = 'time [1]';
    end
    
    t = Signal.Times(fa.p.parameters.value.sampling_frequency, fa.p.nsamples.value, fa.p.time_scale.value,fa.p.t0.value);

    n_subplots = double( int16(plot_re)+int16(plot_im)+int16(plot_mg) );
    i_subplot = 0;
    
    if plot_re
        i_subplot = i_subplot+1;
        subplot(n_subplots,1,i_subplot)    
        if plot_sg
            plot( t, real( signal ) );
            hold on
        end
        for i=1:nterms_
            plot( t, real( terms(:,i) ) );
            hold on
        end
        legend(legend_{:});
        xlabel(xlabel_)
        ylabel('Re')
        hold off
    end
    
    if plot_im
        i_subplot = i_subplot+1;
        subplot(n_subplots,1,i_subplot)
        if plot_sg
            plot( t, imag( signal ) );
            hold on
        end
        for i=1:nterms_
            plot( t, imag( terms(:,i) ) );
            hold on
        end    
        legend(legend_{:});
        xlabel(xlabel_)
        ylabel('Im')
        hold off
    end
    
    if plot_mg
        i_subplot = i_subplot+1;
        subplot(n_subplots,1,i_subplot)
        if plot_sg
            plot( t, abs( signal ) );
            hold on
        end
        for i=1:nterms_
            plot( t, abs( terms(:,i) ) );
            hold on
        end    
        legend(legend_{:});
        xlabel(xlabel_)
        ylabel('magnitude')
        hold off
    end
end
